from __future__ import print_function

from numpy import pi
import h5py

def LoadAsterixMany(filedescriptors):
    result = [LoadAsterixRawHDF(fd['filename'], friendly_name=fd['friendly_name'])
              for fd in filedescriptors]
    return result

def LoadAsterixRawHDF(filename, path=None, friendly_name="", format="HDF5", **kwargs):
    if path == None:
        path = os.getcwd()
    print("format:", format)
    if friendly_name.endswith('hdf'):
        format = "HDF4"
    else: #h5
        format = "HDF5"

    if format == "HDF4":
        print("converting hdf4 to hdf5")
        (tmp_fd, tmp_path) = tempfile.mkstemp() #temporary file for converting to HDF5
        print(tmp_path)
        print('h4toh5 %s %s' , os.path.join(path, filename), tmp_path)
        subprocess.call(['h4toh5', os.path.join(path, filename), tmp_path])
        hdf_obj = h5py.File(tmp_path, mode='r')
    else:
        hdf_obj = h5py.File(os.path.join(path, filename), mode='r')
    run_title = hdf_obj.keys()[0]
    run_obj = hdf_obj[run_title]
    state = hdf_to_dict(run_obj['ASTERIX'])
    monitor = hdf_to_dict(run_obj['scalars'])
    tof = run_obj['ordela_tof_pz']['tof'].value.astype(float64)
    twotheta_pixel = run_obj['ordela_tof_pz']['X'].value.astype(float64)
    data = run_obj['ordela_tof_pz']['data'].value.astype(float64)
    creation_story = "LoadAsterixRawHDF('{fn}')".format(fn=filename)
    output_objs = []
    #for col in range(4):
    info = [{"name": "tof", "units": "nanoseconds", "values": tof[:-1]},
            {"name": "xpixel", "units": "pixels", "values": twotheta_pixel[:-1]},
            {"name": "Measurements", "cols": [
                {"name": "counts_down_down"},
                {"name": "counts_down_up"},
                {"name": "counts_up_down"},
                {"name": "counts_up_up"},
                {"name": "monitor_down_down"},
                {"name": "monitor_down_up"},
                {"name": "monitor_up_down"},
                {"name": "monitor_up_up"},
                {"name": "pixels"},
                {"name": "count_time"}]},
            {"PolState": '', "filename": filename, "start_datetime": None,
             "state": state, "CreationStory":creation_story, "path":path}]
    data_array = zeros((500, 256, 10))
    data_array[:, :, -2] = 1.0 # pixels

    data_array[:, :, -1] = 1.0 # count time
    for i in range(4):
        data_array[:, :, i] = data[i,:,:]
        data_array[:, :, i+4] = monitor['microamphours_p%d' % (i,)]
        #data_array[:,:,0] = data[col,:,:]
        #output_objs.append(MetaArray(data_array[:], dtype='float', info=info[:]))

    hdf_obj.close()
    if format == "HDF4":
        print("removing temporary file")
        os.remove(tmp_path)
        print("temp file removed")

    #return output_objs
    new_data = MetaArray(data_array[:], dtype='float', info=info[:])
    new_data.friendly_name = friendly_name # goes away on dumps/loads... just for initial object.
    return new_data
def SuperLoadAsterixHDF(filename, friendly_name="", path=None,
                        center_pixel=145.0, wl_over_tof=1.9050372144288577e-5,
                        pixel_width_over_dist=0.0195458*pi/180., format="HDF5"):
    """ loads an Asterix file and does the most common reduction steps,
    giving back a length-4 list of data objects in twotheta-wavelength space,
    with the low-tof region shifted to the high-tof region """
    data_objs = LoadAsterixRawHDF(filename, path, format)
    tth_converted = AsterixPixelsToTwotheta().apply(data_objs, qzero_pixel=center_pixel,
                                                    pw_over_d=pixel_width_over_dist)
    wl_converted = AsterixTOFToWavelength().apply(tth_converted, wl_over_tof=wl_over_tof)
    shifted = AsterixShiftData().apply(wl_converted, edge_bin=180)
    return shifted

def LoadAsterixHDF(filename, friendly_name="", path=None,
                   center_pixel=145.0, wl_over_tof=1.9050372144288577e-5):
    if path == None:
        path = os.getcwd()
    hdf_obj = h5py.File(os.path.join(path, filename))
    run_title = hdf_obj.keys()[0]
    run_obj = hdf_obj[run_title]
    state = hdf_to_dict(run_obj['ASTERIX'])
    print(state)
    monitor = hdf_to_dict(run_obj['scalars'])
    tof = run_obj['ordela_tof_pz']['tof'].value.astype(float64)
    twotheta_pixel = run_obj['ordela_tof_pz']['X'].value.astype(float64)
    data = run_obj['ordela_tof_pz']['data'].value.astype(float64)
    tof_to_wavelength_conversion = (0.019050372144288577 / 1000.)
    wavelength = (tof * wl_over_tof)[:-1]
    # shift by half-bin width to align to center of tof bins!
    wavelength += (tof[1] - tof[0])/2.0 * wl_over_tof
    # (bins appear to be centered)
    shifted_data = empty(data.shape)
    shifted_data[:, :320, :] = data[:, -320:, :]
    shifted_data[:, 320:, :] = data[:, :-320, :]
    shifted_wavelength = zeros(wavelength.shape)
    shifted_wavelength[:320] = wavelength[-320:]
    shifted_wavelength[320:] = wavelength[:-320] + (wavelength[-1] + wavelength[0])
    pixel_width_over_dist = 0.0195458*pi/180.
    twotheta_offset = float(state['A[0]'])
    twotheta = arctan2((twotheta_pixel - center_pixel) * pixel_width_over_dist, 1.0) * 180./pi + twotheta_offset
    print('tth:', float(state['A[0]']))
    pol_states = {0:'--', 1:'-+', 2:'+-', 3:'++'}
    creation_story = "LoadAsterixHDF('{fn}')".format(fn=filename)
    #wavelength_axis = data_in[:,0,0]
    #twotheta_axis = data_in[0,:,1]
    output_objs = []
    for col in range(4):
        info = [{"name": "wavelength", "units": "Angstroms", "values": shifted_wavelength},
                {"name": "twotheta", "units": "degrees", "values": twotheta},
                {"name": "Measurements", "cols": [
                    {"name": "counts"},
                    {"name": "pixels"},
                    {"name": "monitor"},
                    {"name": "count_time"}]},
                {"PolState": pol_states[col], "filename": filename, "start_datetime": None,
                 "theta": float(state['A[1]']), "det_angle": float(state['A[0]']),
                 "CreationStory":creation_story, "path":path}]
        data_array = zeros((500, 256, 4))
        data_array[:, :, 1] = 1.0 # pixels
        data_array[:, :, 2] = 1.0 # monitor
        data_array[:, :, 3] = 1.0 # count time
        data_array[:, :, 0] = shifted_data[col, :, :]
        output_objs.append(MetaArray(data_array[:], dtype='float', info=info[:]))
    return output_objs

def LoadAsterixData(filename, friendly_name="", path = None):
    if path == None:
        path = os.getcwd()
    pol_states = {2:'--', 3:'-+', 4:'+-', 5:'++'}
    creation_story = "LoadAsterixData('{fn}')".format(fn=filename)
    data_in = loadtxt(os.path.join(path, filename)).reshape(500, 256, 6)
    wavelength_axis = data_in[:, 0, 0]
    twotheta_axis = data_in[0, :, 1]
    output_objs = []
    for col in range(2, 6):
        info = [{"name": "wavelength", "units": "Angstroms", "values": wavelength_axis},
                {"name": "twotheta", "units": "degrees", "values": twotheta_axis},
                {"name": "Measurements", "cols": [
                    {"name": "counts"},
                    {"name": "pixels"},
                    {"name": "monitor"},
                    {"name": "count_time"}]},
                {"PolState": pol_states[col], "filename": filename, "start_datetime": None,
                 "CreationStory":creation_story, "path":path}]
        data_array = zeros((500, 256, 4))
        data_array[:, :, 1] = 1.0 # pixels
        data_array[:, :, 2] = 1.0 # monitor
        data_array[:, :, 3] = 1.0 # count time
        data_array[:, :, 0] = data_in[:, :, col]
        output_objs.append(MetaArray(data_array[:], dtype='float', info=info[:]))
    return output_objs

def LoadAsterixSpectrum(filename, friendly_name="", path=None):
    spec = LoadText(filename, path, first_as_x=True)
    spec._info[0]["name"] = "tof"
    spec._info[0]["units"] = "microseconds"
    if spec.shape[1] == 1:
        spec._info[1]['cols'][0]['name'] = 'spectrum'
    elif spec.shape[1] == 4:
        for i, pol in enumerate(Filter2D.polarizations):
            spec._info[1]['cols'][i]['name'] = 'spectrum_%' % (pol,)
    return spec



# This program is public domain
# -*- coding: UTF-8 -*-
"""
Load a Bruker raw file into a reflectometry data structure.
"""
from __future__ import print_function

import os
import datetime
import time
import logging
import traceback
from io import BytesIO

import numpy as np

from reductus.dataflow.lib import unit
from reductus.dataflow.lib import iso8601

from . import bruker
from . import rigaku
from . import xrdml
from . import refldata
from .resolution import FWHM2sigma


def load_entries(filename, file_obj=None, entries=None):
    """
    Load the entries from an X-ray file.
    """
    #logging.info("loading X-ray file " + filename)
    if filename.endswith('.ras'):
        return load_rigaku_entries(filename, file_obj)
    elif filename.endswith('.xrdml'):
        return load_xrdml_entries(filename, file_obj)
    else:
        return load_bruker_entries(filename, file_obj)


def load_bruker_entries(filename, file_obj=None):
    """
    Load the entries from a Bruker file.
    """
    if file_obj is None:
        bruker_file = bruker.load(filename)
    else:
        bruker_file = bruker.loads(file_obj.read())

    entries = [BrukerRefl(bruker_file, entryid, filename)
               for entryid, _ in enumerate(bruker_file['data'])]
    return entries

def load_rigaku_entries(filename, file_obj=None):
    """
    Load the entries from a Rigaku file.
    """
    if file_obj is None:
        datasets = rigaku.load(filename)
    else:
        datasets = rigaku.loads(file_obj.read())
    joined_data = rigaku.join(datasets)

    entries = [RigakuRefl(joined_data, filename)]
    return entries

def load_xrdml_entries(filename, file_obj=None):
    """
    Load the entries from an XRDML file.
    """
    if file_obj is None:
        xrdml_file = xrdml.load(filename)
    else:
        xrdml_file = xrdml.load(file_obj)

    entries = [XRDMLRefl(entry, entryid, filename)
               for entryid, entry in enumerate(xrdml_file)]
    return entries


TRAJECTORY_INTENTS = {
    'locked coupled': 'specular',
    'unlocked coupled': 'specular',
    'detector scan': 'rock detector',
    'rocking curve': 'rock sample',
    'phi scan': 'rock sample'
}

class BrukerRefl(refldata.ReflData):
    """
    Bruker raw reflectometry file.

    See :class:`refldata.ReflData` for details.
    """
    format = "BrukerRaw"

    def __init__(self, bruker_file, entryid, filename):
        super(BrukerRefl, self).__init__()
        self.entry = "E"+str(entryid+1)  # 1-origin entry number for each entry
        self.path = os.path.abspath(filename)
        self.name = os.path.basename(filename).split('.')[0]
        self.filenumber = self.name # since we don't have a real filenumber
        #import pprint; pprint.pprint(entry)
        self._set_metadata(bruker_file)
        self._set_data(bruker_file['data'][entryid])

    def _set_metadata(self, entry):
        #print(entry['instrument'].values())
        self.probe = "x-rays"

        # parse date into datetime object
        month, day, year = [int(v) for v in entry['date'].split('/')]
        hour, minute, second = [int(v) for v in entry['time'].split(':')]
        self.date = datetime.datetime(year+2000, month, day, hour, minute, second)

        self.description = entry['comment']
        self.instrument = 'BrukerXray'
        # 1-sigma angular resolution (degrees) reported by Bruker.  Our own
        # measurements give a similar value (~ 0.033 degrees FWHM)
        nominal_resolution = 0.03
        self.angular_resolution = FWHM2sigma(nominal_resolution)
        self.slit1.distance = -275.5
        self.slit2.distance = -192.7
        self.slit3.distance = +175.0
        # Set the slit openings so that divergence calculation will yield
        # the nominal resolution for the given slit positions.
        # In practice the xray resolution is defined by optical elements
        # such as Goebel mirrors, and so the slit geometry should not be
        # used to compute resolution.  Slit openings could be used to estimate
        # beam footprint on the sample if we knew the true value...
        nominal_slits = (np.radians(nominal_resolution)
                         * abs(self.slit1.distance - self.slit2.distance))
        self.slit1.x = self.slit2.x = nominal_slits
        self.slit1.y = self.slit2.y = 20.0 # Note: back slits unused in reduction
        #self.slit4.distance = data_as(entry,'instrument/predetector_slit2/distance','mm')
        #self.detector.distance = data_as(entry,'instrument/detector/distance','mm')
        #self.detector.rotation = data_as(entry,'instrument/detector/rotation','degree')
        # Assume that Kalpha1 and Kalpha2 are both present in 2:1 ratio;
        # Normally we should check the value of fixed_inc_monochromator to
        # determine the wavelength and resolution, but the instrument at the
        # NCNR only has a Göbel mirror.
        self.detector.wavelength = entry['alpha_average']
        if self.detector.wavelength == 0.0:
            # this happens on some converted files...
            if 'alpha_1' in entry and 'alpha_2' in entry and 'alpha_21' in entry:
                self.detector.wavelength = (entry['alpha_1'] + entry['alpha_21'] * entry['alpha_2']) / (1.0 + entry['alpha_21'])
                
        self.detector.wavelength_resolution = np.std([
            entry['alpha_1'], entry['alpha_1'], entry['alpha_2']], ddof=1)
        #print("res:", self.angular_resolution, self.detector.wavelength_resolution, self.detector.wavelength)
        self.detector.deadtime = np.array([0.0]) # sorta true.
        self.detector.deadtime_error = np.array([0.0]) # also kinda true.

        self.sample.name = entry['samplename']
        self.sample.description = ""
        self.monitor.base = 'TIME'
        self.monitor.time_step = 0.001  # assume 1 ms accuracy on reported clock
        self.monitor.deadtime = 0.0
        self.polarization = "unpolarized"

    def _set_data(self, data):
        #for k,v in sorted(data.items()): print(k, v)
        raw_intent = bruker.SCAN_TYPE.get(data['scan_type'], "")
        attenuator_state = data['detslit_code'].strip().lower()
        self.intent = TRAJECTORY_INTENTS.get(raw_intent, refldata.Intent.none)
        self.detector.counts = data['values']['count']
        self.detector.counts_variance = self.detector.counts.copy()
        if attenuator_state == 'in':
            ATTENUATOR = 100.
            #self.v = self.detector.counts*ATTENUATOR
            #self.dv = np.sqrt(self.detector.counts_variance) * ATTENUATOR
            self.detector.counts *= ATTENUATOR
            self.detector.counts_variance *= ATTENUATOR**2
        self.detector.dims = self.detector.counts.shape
        n = self.detector.dims[0]
        self.points = n
        self.monitor.counts = np.zeros_like(self.detector.counts)
        self.monitor.counts_variance = np.zeros_like(self.detector.counts)
        self.monitor.count_time = np.ones_like(self.detector.counts) * data['step_time']
        if raw_intent in ["locked coupled", "unlocked coupled"]:
            self.sample.angle_x = data['theta_start'] + np.arange(n, dtype='float') * data['increment_1'] / 2.0
            self.detector.angle_x = data['two_theta_start'] + np.arange(n, dtype='float') * data['increment_1']
            self.sample.angle_x_target = self.sample.angle_x
            self.detector.angle_x_target = self.detector.angle_x
            self.scan_value = [self.sample.angle_x, self.detector.angle_x]
            self.scan_units = ['degrees', 'degrees']
            self.scan_label = ['theta', 'two_theta']
        elif raw_intent == 'detector scan':
            self.sample.angle_x = data['theta_start']
            self.detector.angle_x = data['two_theta_start'] + np.arange(n, dtype='float') * data['increment_1']
            self.sample.angle_x_target = self.sample.angle_x
            self.detector.angle_x_target = self.detector.angle_x
            self.scan_value = [self.detector.angle_x]
            self.scan_units = ['degrees']
            self.scan_label = ['two_theta']
        elif raw_intent in ['rocking curve', 'phi scan']:
            # this may not be right at all.  I can't understand what reflred/loadraw.tcl is doing here
            self.sample.angle_x = data['theta_start'] - np.arange(n, dtype='float') * data['increment_1']
            self.detector.angle_x = data['two_theta_start'] + np.arange(n, dtype='float') * data['increment_1']
            self.sample.angle_x_target = self.sample.angle_x
            self.detector.angle_x_target = self.detector.angle_x
            self.scan_value = [self.sample.angle_x, self.detector.angle_x]
            self.scan_units = ['degrees', 'degrees']
            self.scan_label = ['theta', 'two_theta']
        else:
            raise ValueError("Unknown sample angle in file")
        self.Qz_target = np.nan

class RigakuRefl(refldata.ReflData):
    """
    Bruker raw reflectometry file.

    See :class:`refldata.ReflData` for details.
    """
    format = "BrukerRaw"

    def __init__(self, dataset, filename):
        super(RigakuRefl, self).__init__()
        self.entry = "entry"
        self.path = os.path.abspath(filename)
        self.name = os.path.basename(filename).split('.')[0]
        self.filenumber = self.name # since we don't have a real filenumber
        #import pprint; pprint.pprint(entry)
        self._set_metadata(dataset)
        self._set_data(dataset)

    def _set_metadata(self, entry):
        #print(entry['instrument'].values())
        self.probe = "x-rays"

        # parse date into datetime object
        self.date = datetime.datetime.fromtimestamp(time.mktime(entry['start_time']))

        self.description = entry['comment']
        self.instrument = 'RigakuXray'
        self.detector.deadtime = np.array([0.0]) # sorta true.
        self.detector.deadtime_error = np.array([0.0]) # also kinda true.

        self.sample.name = entry['sample']
        self.sample.description = ""
        self.monitor.base = 'TIME'
        self.monitor.time_step = 0.001  # assume 1 ms accuracy on reported clock
        self.monitor.deadtime = 0.0
        self.polarization = "unpolarized"

    def _set_data(self, data):
        # Resolution info (returned as 1-sigma from rigaku reader)
        self.angular_resolution = data['angular_divergence']
        self.slit1.distance = data['slit1_distance']
        self.slit2.distance = data['slit2_distance']
        self.slit3.distance = data['slit3_distance']
        self.slit4.distance = data['slit4_distance']
        self.slit1.x = data['axis']['IncidentSlitBox'][2]
        self.slit2.x = data['axis']['IncidentAxdSlit'][2]
        self.slit3.x = data['axis']['ReceivingSlitBox1'][2]
        self.slit4.x = data['axis']['ReceivingSlitBox2'][2]

        self.detector.counts = data['y']
        self.detector.counts_variance = data['y_err']**2
        self.detector.dims = self.detector.counts.shape
        attenuation = data['axis']['Attenuator'][2]
        if attenuation != 0. and attenuation != 1.:
            self.detector.counts /= attenuation
            self.detector.counts_variance /= attenuation**2
        n = self.detector.dims[0]
        self.points = n
        self.detector.wavelength = np.ones(n) * data['wavelength']
        self.detector.wavelength_resolution = np.ones(n) * data['wavelength_resolution']
        self.monitor.counts = np.zeros(n)
        self.monitor.counts_variance = np.zeros(n)
        self.monitor.count_time = data['count_time']
        sample = data['axis']['Omega'][2]
        detector = data['axis']['TwoTheta'][2]
        offset = sample - detector/2
        scan_axis = data['scan_axis']
        if scan_axis in ["TwoThetaOmega", "TwoThetaTheta"]:
            self.sample.angle_x = data['x']/2 + offset
            self.detector.angle_x = data['x']
            self.scan_value = [self.sample.angle_x, self.detector.angle_x]
            self.scan_units = ['degrees', 'degrees']
            self.scan_label = ['theta', 'two_theta']
            if offset > 0:
                self.intent = refldata.Intent.backp
            elif offset < 0:
                self.intent = refldata.Intent.backm
            else:
                self.intent = refldata.Intent.spec
        elif scan_axis == 'Omega':
            self.sample.angle_x = data['x']
            self.detector.angle_x = np.ones(n) * detector
            self.scan_value = [self.sample.angle_x]
            self.scan_units = ['degrees']
            self.scan_label = ['theta']
            self.intent = refldata.Intent.rock3
        elif scan_axis == 'TwoTheta':
            self.sample.angle_x = np.ones(n) * sample
            self.detector.angle_x = data['x']
            self.scan_value = [self.sample.angle_x]
            self.scan_units = ['degrees']
            self.scan_label = ['theta']
            self.intent = refldata.Intent.rock4
        else:
            raise ValueError("Unknown scan type " + scan_axis)
        self.sample.angle_x_target = self.sample.angle_x
        self.detector.angle_x_target = self.detector.angle_x
        self.Qz_target = np.nan


class XRDMLRefl(refldata.ReflData):
    """
    Malvern Panalytical (.xrdml) raw reflectometry file.

    See :class:`refldata.ReflData` for details.
    """
    format = "XRDMLRaw"

    def __init__(self, xrdml_file, entryid, filename):
        super(XRDMLRefl, self).__init__()
        self.entry = "E"+str(entryid+1)  # 1-origin entry number for each entry
        self.path = os.path.abspath(filename)
        self.name = os.path.basename(filename).split('.')[0]
        self.filenumber = self.name # since we don't have a real filenumber
        #import pprint; pprint.pprint(entry)
        self._set_metadata(xrdml_file)
        self._set_data(xrdml_file)

    def _set_metadata(self, entry):
        #print(entry['instrument'].values())
        self.probe = "x-rays"

        # parse date into datetime object
        self.date = datetime.datetime.fromisoformat(entry["startTimeStamp"])

        self.description = entry['comment']
        self.instrument = 'Malvern Panalytical'
        # 1-sigma angular resolution (degrees) reported by Bruker.  Our own
        # measurements give a similar value (~ 0.033 degrees FWHM)
        #nominal_resolution = 0.03
        self.angular_resolution = None
        self.slit1.distance = -entry['incidentRadius']
        self.slit1.x = entry['sourceLineWidth']
        self.slit2.distance = -entry['divergenceSlitDistance']
        self.slit2.x = entry['divergenceSlitHeight']
        self.slit3.distance = entry['diffractedRadius']
        self.slit3.x = entry['receivingSlitHeight']

        self.slit1.y = self.slit2.y = 20.0 # Note: back slits unused in reduction

        ratio = entry['ratioKAlpha2KAlpha1']
        self.detector.wavelength = (entry['kAlpha1'] + ratio * entry['kAlpha2']) / (1.0 + ratio)
        # Var(Z)=p(1)σ_1^2+ p(2)σ_2^2+p(1)p(2)(μ1−μ2)2
        # from table II in https://journals.aps.org/pra/pdf/10.1103/PhysRevA.37.2404
        # width_kAlpha1 = FWHM2sigma(0.000461)
        # width_kAlpha2 = FWHM2sigma(0.00061)
        # probability of kAlpha2:
        p = ratio / (1 + ratio)
        # self.detector.wavelength_resolution = np.sqrt((1-p) * width_kAlpha1**2 + p * width_kAlpha2**2 + (1-p) * p * (entry['kAlpha1'] - entry['kAlpha2'])**2)
        # the width of the line seems to make no significant contribution...
        self.detector.wavelength_resolution = np.sqrt((1-p) * p * (entry['kAlpha1'] - entry['kAlpha2'])**2)

        #print("res:", self.angular_resolution, self.detector.wavelength_resolution, self.detector.wavelength)
        self.detector.deadtime = np.array([0.0]) # sorta true.
        self.detector.deadtime_error = np.array([0.0]) # also kinda true.

        self.sample.name = entry['sample_id']
        self.sample.description = entry['sample_name']
        self.monitor.base = 'TIME'
        self.monitor.time_step = 0.001  # assume 1 ms accuracy on reported clock
        self.monitor.deadtime = 0.0
        self.polarization = "unpolarized"

    def _set_data(self, data):
        #for k,v in sorted(data.items()): print(k, v)
        intents = {
            'Phi': 'rock sample',
            'Omega': 'rock sample',
            '2Theta': 'rock detector',
        }

        scanAxis = data['scanAxis']
        self.intent = intents.get(scanAxis, 'specular')

        attenuation = data['attenuation']
        counts = data['counts']
        self.detector.counts = counts * attenuation
        self.detector.counts_variance = counts * attenuation**2
        self.detector.dims = self.detector.counts.shape
        self.attenuation = attenuation
        n = self.detector.dims[0]
        self.points = n
        self.monitor.counts = np.zeros_like(self.detector.counts)
        self.monitor.counts_variance = np.zeros_like(self.detector.counts)
        self.monitor.count_time = np.ones_like(self.detector.counts) * data['count_time']
        if scanAxis == 'Phi':
            self.sample.angle_x = data['Phi']
            self.detector.angle_x = -data['Phi']
        else:
            self.sample.angle_x = data['Omega']
            self.detector.angle_x = data['2Theta']

        self.sample.angle_x_target = self.sample.angle_x
        self.detector.angle_x_target = self.detector.angle_x
        self.scan_value = [self.sample.angle_x, self.detector.angle_x]
        self.scan_units = ['degrees', 'degrees']
        self.scan_label = ['theta', 'two_theta']
        self.Qz_target = np.nan

def demo():
    import sys
    from .load import setup_fetch, fetch_uri
    from .scale import apply_norm
    from .steps import divergence
    if len(sys.argv) == 1:
        print("usage: python -m reflred.xrawref file...")
        sys.exit(1)
    setup_fetch()
    plotted_datasets = 0
    for uri in sys.argv[1:]:
        try:
            entries = fetch_uri(uri, loader=load_entries)
        except Exception as exc:
            print("Error while loading", uri, ':', str(exc))
            #traceback.print_exc(); raise
            continue

        # print the first entry
        #print(entries[0])

        # plot all the entries
        #pylab.figure()
        for entry in entries:
            entry = divergence(entry)
            apply_norm(entry, base='time')
            entry.plot()
            plotted_datasets += 1

    if plotted_datasets:
        import pylab
        pylab.legend()
        pylab.show()
    else:
        print("no data to plot")

if __name__ == "__main__":
    demo()



# This program is public domain
"""
Error propagation algorithms for linear approximation to error.

This is a set of functions for processing uncertainties on arrays,
with both in-place and returned options.  See the :mod:`uncertainty`
module for an object-oriented wrapper.

Warning: like the underlying numpy library, the inplace operations
may return values of the wrong type if some of the arguments are
integers, so be sure to create them with floating point inputs.

Linear error approximation relies on linearity and independence.
These conditions may be broken in some circumstances.   The
*uncertainties* package on pypi does an excellent job of propagating
error through expressions, including those where the variable is used
multiple times.  Unfortunately, it is slow for large arrays, and so
this package is still needed.
"""
from __future__ import division  # Get true division

import numpy as np

# TODO: use numba for faster vector processing
# Most operations can be done in-place with one pass through the data saving
# both time and memory while being simpler to code.

def mean(X, varX, biased=True, axis=None, dtype=None, out=None, keepdims=False):
    # type: (np.ndarray, np.ndarray, bool) -> (float, float)
    r"""
    Return the mean and variance of a dataset.

    If varX is estimated from the data, then use *biased=True* so that the
    estimated variance is scaled by the normalized $\chi^2$.  See the
    wikipedia page for the weighted arithmetic mean for details.
    """
    # Use preallocated output (e.g., within an array slice) if provided
    Mout, varMout = (None, None) if out is None else out
    total_weight = np.sum(1./varX, axis=axis, dtype=dtype, keepdims=keepdims)
    M = np.sum(X/varX, axis=axis, dtype=dtype, out=Mout, keepdims=keepdims)
    if varMout is None:
        varM = 1./total_weight
    else:
        varM = varMout
        varM.fill(1.)
        varM /= total_weight
    if biased:
        # Scale by normalized chisq if variance calculation is biased
        chisq = np.sum((X-M)**2/varX, axis=axis, dtype=dtype, keepdims=keepdims)
        dof = np.prod(X.shape)/np.prod(chisq.shape) - 1
        varM *= chisq/dof
    return M, varM


def sum(X, varX, axis=None, dtype=None, out=None, keepdims=False):
    # type: (np.ndarray, np.ndarray, ...) -> (float, float)
    r"""
    Return the sum and variance of a dataset.

    Follows the numpy sum interface, except a pair of output arrays is required
    if you want to reuse an output.
    """
    Mout, varMout = (None, None) if out is None else out
    M = np.sum(X, axis=axis, dtype=dtype, out=Mout, keepdims=keepdims)
    varM = np.sum(varX, axis=axis, dtype=dtype, out=varMout, keepdims=keepdims)
    return M, varM


def cumsum(X, varX, axis=None, dtype=None, out=None):
    # type: (np.ndarray, np.ndarray, ...) -> (float, float)
    r"""
    Return the cumulative sum and variance of a dataset.

    Follows the numpy cumsum interface, except a pair of output arrays is
    required if you want to reuse an output.
    """
    Mout, varMout = (None, None) if out is None else out
    M = np.cumsum(X, axis=axis, dtype=dtype, out=Mout)
    varM = np.cumsum(varX, axis=axis, dtype=dtype, out=varMout)
    return M, varM


def average(X, varX, W, varW, axis=None):
    # type: (np.ndarray, np.ndarray, np.ndarray, np.ndarray, int) -> (float, float)
    r"""
    Return the weighted average of a dataset, with uncertainty in the weights.

    This is usual weighted average formula with gaussian uncertainty
    propagation. Use W=1, varW=0 for the simple gaussian average.

    Results checked against Monte Carlo simulation in explore/gaussian_average.py
    """
    # TODO: make this work for nD arrays when axis somewhere in the middle.
    # This requires keepdims for Swx, Sw, and then remove dim at the end
    Swx = np.sum(X*W, axis=axis)
    # Note: only need to check for W as scalar. The X in the expression of
    # varM will automatically promote all other scalars to vectors as needed.
    if np.isscalar(W):
        Sw = W*(np.prod(X.shape) if axis is None else X.shape[axis])
    else:
        Sw = np.sum(W, axis=axis)
    M = Swx/Sw
    varM = np.sum((W/Sw)**2*varX + ((X*Sw - Swx)/Sw**2)**2*varW, axis=axis)
    return M, varM


def interp(X, Xp, Fp, varFp, left=None, right=None):
    # type: (np.ndarray, np.ndarray, np.ndarray, np.ndarray, int) -> (np.ndarray, np.ndarray)
    """
    Linear interpolation of x points into points (xk,fk +/- dfk).

    xp is assumed to be monotonically increasing.  The interpolated
    value is undefined at duplicate points.

    Fp, varFp can be n-dimensional.

    left is (value,variance) to use for points before the range of xp, or
    None for the initial value.

    right is (value,variance) to use for points after the range of xp, or
    None for the final value.
    """
    idx = np.searchsorted(Xp[1:-1], X, side="right")
    if len(Xp) > 1:
        # Support repeated values in Xp, which will lead to 0/0 errors if the
        # interpolated point is one of the repeated values.
        p = (Xp[idx+1]-X)/(Xp[idx+1]-Xp[idx])
        if Fp.ndim > 1: # n-D support requires manual broadcast
            extra_dims = (np.newaxis,)*(Fp.ndim-1)
            p = p[(..., *extra_dims)]
        F = p*Fp[idx] + (1-p)*Fp[idx+1]
        # simple propagation of error formula for calculation of F, confirmed
        # by monte carlo simulation.
        varF = p**2*varFp[idx] + (1-p)**2*varFp[idx+1]
    else:
        F, varF = Fp[idx], varFp[idx]
    #print p,F,varF,idx
    if left is None:
        left = Fp[0], varFp[0]
    if right is None:
        right = Fp[-1], varFp[-1]
    F[X < Xp[0]], varF[X < Xp[0]] = left
    F[X > Xp[-1]], varF[X > Xp[-1]] = right

    return F, varF


def div(X, varX, Y, varY):
    """Division with error propagation"""
    # Direct algorithm:
    #   Z = X/Y
    #   varZ = (varX/X**2 + varY/Y**2) * Z**2
    #        = (varX + varY * Z**2) / Y**2
    # Indirect algorithm to minimize intermediates
    #np.seterr(all='raise')
    Z = X/Y      # truediv => Z is a float
    varZ = Z**2  # Z is a float => varZ is a float
    varZ *= varY
    varZ += varX
    # Z/Y/Y is more expensive than Z/Y**2 (poor data locality going through
    # the Y array twice), but it avoids creating a temporary for Y**2.
    # Also, happens to avoid integer overflow on Y**2...
    varZ /= Y
    varZ /= Y
    return Z, varZ


def mul(X, varX, Y, varY):
    """Multiplication with error propagation"""
    # Direct algorithm:
    Z = X * Y
    varZ = Y**2 * varX + X**2 * varY
    # Indirect algorithm won't ensure floating point results
    #   varZ = Y**2
    #   varZ *= varX
    #   Z = X**2   # Using Z to hold the temporary
    #   Z *= varY
    #   varZ += Z
    #   Z[:] = X
    #   Z *= Y
    return Z, varZ


def sub(X, varX, Y, varY):
    """Subtraction with error propagation"""
    Z = X - Y
    varZ = varX + varY
    return Z, varZ


def add(X, varX, Y, varY):
    """Addition with error propagation"""
    Z = X + Y
    varZ = varX + varY
    return Z, varZ

def sqrt(X, varX):
    """Square root with error propagation"""
    # Direct algorithm
    #   Z = X**n = sqrt(X)
    #   varZ = n*n * varX/X**2 * Z**2 = varX/(4*X)
    # Indirect algorithm to minimize intermediates
    Z = np.sqrt(X)
    varZ = varX/X
    varZ /= 4
    return Z, varZ

def exp(X, varX):
    """Exponentiation with error propagation"""
    Z = np.exp(X)
    varZ = varX * Z**2
    return Z, varZ


def log(X, varX):
    """Logarithm with error propagation"""
    Z = np.log(X)
    varZ = varX / X**2
    return Z, varZ


def sin(X, varX):
    return np.sin(X), varX*np.cos(X)**2


def cos(X, varX):
    return np.cos(X), varX*np.sin(X)**2


def tan(X, varX):
    return np.tan(X), varX/np.cos(X)**2


def arcsin(X, varX):
    return np.arcsin(X), varX/(1-X**2)


def arccos(X, varX):
    return np.arccos(X), varX/(1-X**2)


def arctan(X, varX):
    return np.arctan(X), varX/(1+X**2)**2


def arctan2(Y, varY, X, varX):
    # df/dX = Y/(X**2 + Y**2)
    # df/dY = X/(X**2 + Y**2)
    # varZ = (df/dX)**2 * varX + (df/dY)**2 * varY
    Z = np.arctan2(Y, X)
    varZ = (Y**2 * varX + X**2 * varY) / (X**2 + Y**2)**2
    return Z, varZ


# Confirm this formula before using it
# def pow(X,varX, Y,varY):
#    Z = X**Y
#    varZ = (Y**2 * varX/X**2 + varY * numpy.log(X)**2) * Z**2
#    return Z,varZ
#

def pow(X, varX, n):
    """X**n with error propagation"""
    # Direct algorithm
    #   Z = X**n
    #   varZ = n*n * varX/X**2 * Z**2
    # Indirect algorithm to minimize intermediates
    Z = X**n
    varZ = varX/X
    varZ /= X
    varZ *= Z
    varZ *= Z
    varZ *= n**2
    return Z, varZ


def pow2(X, varX, Y, varY):
    """X**Y with error propagation"""
    # Direct algorithm
    #   Z = X**Y
    #   varZ = Z**2 * ((Y*varX/X)**2 + (log(X)*varY)**2)
    Z = X**Y
    varZ = varX/X
    varZ *= Y
    varZ *= varZ
    T = np.log(X)
    T *= varY
    T *= T
    varZ += T
    del T
    varZ *= Z
    varZ *= Z
    return Z, varZ


def div_inplace(X, varX, Y, varY):
    """In-place division with error propagation"""
    # Z = X/Y
    # varZ = (varX + varY * (X/Y)**2) / Y**2 = (varX + varY * Z**2) / Y**2
    X /= Y     # X now has Z = X/Y
    T = X**2   # create T with Z**2
    T *= varY  # T now has varY * Z**2
    varX += T  # varX now has varX + varY*Z**2
    del T      # may want to use T[:] = Y for vectors
    T = Y**2   # reuse T for Y**2
    varX /= T  # varX now has varZ
    return X, varX


def mul_inplace(X, varX, Y, varY):
    """In-place multiplication with error propagation"""
    # Z = X * Y
    # varZ = Y**2 * varX + X**2 * varY
    T = Y**2   # create T with Y**2
    varX *= T  # varX now has Y**2 * varX
    del T      # may want to use T[:] = X for vectors
    T = X**2   # reuse T for X**2 * varY
    T *= varY  # T now has X**2 * varY
    varX += T  # varX now has varZ = X**2*varY + Y**2*varX
    X *= Y     # X now has Z = X*Y
    return X, varX


def sub_inplace(X, varX, Y, varY):
    """In-place subtraction with error propagation"""
    # Z = X - Y
    # varZ = varX + varY
    X -= Y
    varX += varY
    return X, varX


def add_inplace(X, varX, Y, varY):
    """In-place addition with error propagation"""
    # Z = X + Y
    # varZ = varX + varY
    X += Y
    varX += varY
    return X, varX


def pow_inplace(X, varX, n):
    """In-place X**n with error propagation"""
    # Direct algorithm
    #   Z = X**n
    #   varZ = abs(n) * varX/X**2 * Z**2
    # Indirect algorithm to minimize intermediates
    varX /= X
    varX /= X     # varX now has varX/X**2
    X **= n       # X now has Z = X**n
    varX *= X
    varX *= X     # varX now has varX/X**2 * Z**2
    varX *= n**2  # varX now has varZ
    return X, varX


def pow2_inplace(X, varX, Y, varY):
    """In-place X**Y with error propagation"""
    # Direct algorithm
    #   Z = X**Y
    #   varZ = Z**2 * ((Y*varX/X)**2 + (log(X)*varY)**2)
    varX /= X
    varX *= Y
    varX *= varX # varX now has (Y*varX/X)**2
    T = np.log(X)
    T *= varY
    T *= T
    varX += T # varX now has (Y*varX/X)**2 + (log(X)*varY)**2
    del T
    X **= Y   # X now has Z = X**Y
    varX *= X
    varX *= X # varX now has varZ
    return X, varX

def test():
    X, varX = 2, 0.04
    Y, varY = 3, 0.09
    N = 3
    def _check(op, result, variance):
        if op is pow:
            Z, varZ = op(X, varX, N)
        else:
            Z, varZ = op(X, varX, Y, varY)
        assert abs(Z-result)/result < 1e-13 and (varZ-variance)/variance < 1e-13, \
            "expected (%g,%g) got (%g,%g)"%(result, variance, Z, varZ)

    _check(add, X+Y, varX + varY)
    _check(sub, X-Y, varX + varY)
    _check(mul, X*Y, Y**2*varX + X**2*varY)
    _check(div, X/Y, (Y**2*varX + X**2*varY)/Y**4)
    _check(pow, X**N, varX/X**2 * X**(2*N) * N**2)
    _check(pow2, X**Y, X**(2*Y) * ((Y*varX/X)**2 + (np.log(X)*varY)**2))

def test_against_uncertainties_package():
    try:
        from uncertainties import ufloat
        from uncertainties import umath
        from math import sqrt as math_sqrt
    except ImportError:
        return
    X, varX = 0.5, 0.04
    Y, varY = 3, 0.09
    N = 3
    ux = ufloat(X, math_sqrt(varX))
    uy = ufloat(Y, math_sqrt(varY))

    def _compare(result, u):
        Z, varZ = result
        assert abs(Z-u.n)/u.n < 1e-13 and (varZ-u.s**2)/u.s**2 < 1e-13, \
            "expected (%g,%g) got (%g,%g)"%(u.n, u.s**2, Z, varZ)

    def _check_pow(u):
        _compare(pow(X, varX, N), u)

    def _check_unary(op, u):
        _compare(op(X, varX), u)

    def _check_binary(op, u):
        _compare(op(X, varX, Y, varY), u)

    _check_pow(ux**N)
    _check_binary(add, ux+uy)
    _check_binary(sub, ux-uy)
    _check_binary(mul, ux*uy)
    _check_binary(div, ux/uy)
    _check_binary(pow2, ux**uy)

    _check_unary(exp, umath.exp(ux))
    _check_unary(log, umath.log(ux))
    _check_unary(sqrt, umath.sqrt(ux))
    _check_unary(sin, umath.sin(ux))
    _check_unary(cos, umath.cos(ux))
    _check_unary(tan, umath.tan(ux))
    _check_unary(arcsin, umath.asin(ux))
    _check_unary(arccos, umath.acos(ux))
    _check_unary(arctan, umath.atan(ux))
    _check_binary(arctan2, umath.atan2(ux, uy))

def test_average():
    def _compare(result, u):
        Z, varZ = result
        U, varU = u
        assert abs(Z-U)/U < 1e-13 and (varZ-varU)/varU < 1e-13, \
            "expected (%g,%g) got (%g,%g)"%(U, varU, Z, varZ)
    x = np.array([1, 2, 3])
    xvar = np.array([0.2, 0.1, 0.05])
    xvar_const = np.array([0.2, 0.2, 0.2])
    w = np.array([1, 1, 1])
    wvar = np.array([0, 0, 0])
    s, svar = sum(x, xvar)
    # Test simple gaussian average gives the correct result
    _compare(average(x, xvar, 1, 0), (s/3, svar/9))
    # Test that program accepts scalar and vector inputs
    _compare(average(x, xvar_const[0], 1, 0), average(x, xvar_const, w, wvar))

if __name__ == "__main__":
    test()
    test_against_uncertainties_package()
    test_average()



# This program is public domain.
#
# Author: Paul Kienzle
# Based on iso8601 by Michael Twomey
"""
ISO 8601 date time support

Basic usage::

    >>> from reductus.dataflow.lib import iso8601
    >>> iso8601.parse_date("2007-01-25T12:34:56Z")
    datetime.datetime(2007, 1, 25, 12, 34, 56, tzinfo=<TimeZone 'UTC'>)
    >>> iso8601.parse_date("2007-01-25T12:34:56-0500")
    datetime.datetime(2007, 1, 25, 12, 34, 56, tzinfo=<TimeZone '-0500'>)
    >>> iso8601.seconds_since_epoch("2007-01-25T12:00:00Z")
    1169726400.0
    >>> print(iso8601.format_date(1169726400.0)) # doctest: +SKIP
    2007-01-25T07:00:00-0500

The above examples assume US Eastern Standard Time, and may be different
in your time zone.

This code is incomplete.  It does not accept compressed formats (YYYYMMDD
and HHMMSS), week numbers (YYYY-Www-D), day numbers (YYYY-DDD), durations
P#Y#M#DT#H#M#S, intervals (date/date or date/interval), or repeating intervals
(R#/date/date or R#/date/interval).
"""
__all__ = ["parse_date", "format_date", "now", "seconds_since_epoch"]

import time
from datetime import datetime, timedelta, tzinfo
import re

try:
    #from typing import Union
    pass
except ImportError:
    pass

ISO8601_RELAXED = re.compile(r"""^ # anchor to start of string
  (?P<year>[0-9]{4})               # year   YYYY
  (-(?P<month>[0-9]{1,2})          # month  -M or -MM
    (-(?P<day>[0-9]{1,2})          # day    -D or -DD
      (.                           # separator (usually T or space)
        (?P<hour>[0-9]{1,2})       # hour   H or HH
        :(?P<minute>[0-9]{2})      # minute :MM
        (:(?P<second>[0-9]{2})     # second :SS
          (\.(?P<fraction>[0-9]+)  # fractional second .SSS to arbitrary precision
          )?                       # .SSS is optional
        )?                         # SS.SSS is optional
        (?P<timezone>
          Z                        # use Z for UTC
          |
          (?P<tzprefix>[+-])       # +/- offset
          (?P<tzhour>[0-9]{1,2})   # hour offset H or HH
          (:?                      # optional separator for minute offset
            (?P<tzminute>[0-9]{2}) # minute offset MM
          )?                       # optional minute offset
        )?                         # optional time zone
      )?                           # optional time+time zone
    )?                             # YYYY-MM only
  )?                               # YYYY only
  $                                # anchor to end of the string
  """, re.VERBOSE)
ISO8601_STRICT = re.compile(r"""^ # anchor to start of string
  (?P<year>[0-9]{4})              # year   YYYY
  -(?P<month>[0-9]{2})            # month  -MM
  -(?P<day>[0-9]{2})              # day    -DD
  (\ |T)(?P<hour>[0-9]{2})        # hour   THH
  :(?P<minute>[0-9]{2})           # minute :MM
  (:(?P<second>[0-9]{2})          # second :SS
    (\.(?P<fraction>[0-9]+))?     # optional fractional second .SSS to arbitrary precision
  )?                              # optional SS.SSS
  (?P<timezone>
    Z                             # use Z for UTC
    |
    (?P<tzprefix>[+-])            # +/- offset
    (?P<tzhour>[0-9]{2})          # hour offset HH
    :?                            # optional separator for minute offset
    (?P<tzminute>[0-9]{2})        # minute offset MM
  )?                              # time zone is optional, except for really strict
  $                               # anchor to end of the string
  """, re.VERBOSE)

def now(use_microsecond=False):
    """
    Return the current time as an ISO 8601 string.

    Times are recorded in the local time zone, with an offset from UTC.

    If *use_microsecond* then include fractional seconds in the
    returned string.
    """
    return format_date(time.time(), precision=(6 if use_microsecond else 0))

def format_date(timestamp, precision=0):
    # type: (Union[float, datetime, time_struct], int) -> str
    """
    Construct an ISO 8601 time from a timestamp.

    There are several possible sources for *timestamp*.

    - time.time() returns a floating point number of seconds since the
      UNIX epoch of Jan 1, 1970 UTC.

    - time.localtime(time.time()) returns a time tuple for the current
      time using the local time representation.

    - datetime.datetime.now() returns a datetime object for the current
      time using the local time representation, but with no information
      about time zone.

    - iso8601.parse_date(str) returns a datetime object for a previously
      stored time stamp which retains time zone information.

    In the first three cases the formatted date will use the local time
    representation but include the UTC offset for the local time.  The
    fourth case the UTC offset of the time stamp will be preserved in
    formatting.

    If *precision* is given, encode fractional seconds with this many digits
    of precision. This only works if *timestamp* is datetime object or seconds
    since epoch.
    """
    dt = None
    microsecond = 0

    # Try converting from seconds to time_struct
    try:
        microsecond = int(1000000*(timestamp-int(timestamp)))
        timestamp = time.localtime(timestamp)
    except TypeError:
        # Not a floating point timestamp; could be datetime or time_struct
        pass

    # Try converting from datetime to time_struct
    if isinstance(timestamp, datetime):
        microsecond = timestamp.microsecond
        tz = timestamp.utcoffset()
        if tz is not None:
            dt = tz.days*86400 + tz.seconds
        timestamp = timestamp.timetuple()

    # Find time zone offset
    isdst = timestamp.tm_isdst if timestamp.tm_isdst >= 0 else 0
    if dt is None:
        dt = -(time.timezone, time.altzone)[isdst]

    # Do the formatting
    local = time.strftime('%Y-%m-%dT%H:%M:%S', timestamp)
    fraction = ".%0*d"%(precision, microsecond//10**(6-precision)) if precision else ""
    sign = "+" if dt >= 0 else "-"
    offset = "%02d%02d"%(abs(dt)//3600, (abs(dt)%3600)//60)
    return "".join((local, fraction, sign, offset))

class TimeZone(tzinfo):
    """
    Fixed offset time zone in seconds from UTC
    """
    def __init__(self, offset=0, name=None):
        second = int(abs(offset)+0.5)
        if second%60 != 0:
            raise ValueError("offset must be a whole number of minutes")
        if name is None:
            sign = '-' if offset < 0 else '+'
            hour = second//3600
            minute = (second%3600)//60
            name = "%s%02d%02d"%(sign, hour, minute)
        self.__offset = timedelta(seconds=offset)
        self.__name = name

    def utcoffset(self, dt):
        """
        Return offset from UTC.

        *dt* is True for daylight savings time.
        """
        return self.__offset

    def tzname(self, dt):
        """
        Return a description of the time zone.

        This will just be the string offset, or it will be UTC.

        *dt* is True for daylight savings time.
        """
        return self.__name

    def dst(self, dt):
        """
        Return daylight savings time offset.

        We can't tell from the +/- offset timestamp if daylight savings
        is in effect, or if we are just one timezone removed, so DST always
        returns a time delta of 0.

        *dt* is True for daylight savings time.
        """
        return timedelta(0)

    def __repr__(self):
        return "<TimeZone %r>" % self.__name

UTC = TimeZone(name="UTC")
EPOCH = datetime(1970, 1, 1, tzinfo=UTC)

def parse_date(datestring, default_timezone=UTC, strict=False):
    """
    Parses ISO 8601 dates into datetime objects

    The timezone is parsed from the date string. However it is quite common to
    have dates without a timezone (not strictly correct). In this case the
    default timezone specified in default_timezone is used. This is UTC by
    default.

    If strict is True, then only accept YYYY-MM-DD.HH:MM[:SS.SSS][time zone]

    Raises TypeError if not passed a string.
    Raises ValueError if the string is not a valid time stamp.
    """
    #print("parse_date with", datestring)
    try:
        if not strict:
            m = ISO8601_RELAXED.match(datestring)
        else:
            m = ISO8601_STRICT.match(datestring)
    except TypeError:
        # Cruft python 2.x; in 3.x use 'raise TypeError(...) from None'
        exc = TypeError("parse_date expects a string, not %s"%(str(datestring)))
        exc.__cause__ = None
        raise exc
    if not m:
        raise ValueError("Unable to parse date string %r" % datestring)
    groups = m.groupdict()
    year = int(groups["year"])
    month = int(groups["month"]) if groups["month"] else 1
    day = int(groups["day"]) if groups["day"] else 1
    hour = int(groups["hour"]) if groups["hour"] else 0
    minute = int(groups["minute"]) if groups["minute"] else 0
    second = int(groups["second"]) if groups["second"] else 0
    fraction = int(float("0.%s" % groups["fraction"]) * 1e6) if groups["fraction"] else 0
    if groups["timezone"] is None:
        tz = default_timezone
    elif groups["timezone"] == "Z":
        tz = UTC
    else:
        sign = +1 if groups["tzprefix"] == "+" else -1
        dt = (int(groups["tzhour"])*60
              + (int(groups["tzminute"]) if groups["tzminute"] else 0))
        tz = TimeZone(name=groups["timezone"], offset=sign*dt*60)
    return datetime(year, month, day, hour, minute, second, fraction, tz)

def seconds_since_epoch(date, default_timezone=UTC):
    """
    Parse ISO 8601 dates into seconds since epoch.

    *date* can either be a string or a datetime object with timezone specified.
    """
    if isinstance(date, datetime):
        t = date
    else:
        t = parse_date(date, default_timezone=default_timezone)
    dt = t - EPOCH
    return dt.days*86400 + dt.seconds + dt.microseconds*1e-6


# ================= TESTS =================
def _check_date(s, d, strict): # pragma: no cover
    t = parse_date(s)
    assert t-d == timedelta(0), "%r != %s"%(s, d)
    try:
        parse_date(s, strict=True)
        if not strict:
            raise Exception("exception not raised for strict %r"%s)
    except ValueError as exc:
        if strict:
            raise Exception("unexpected exception for strict %r\n  %s"
                            % (s, str(exc)))
def _check_fail(s): # pragma: no cover
    try:
        parse_date(s)
    except Exception:
        return
    raise Exception("exception not raised for %r")

def _check_equal(s1, s2): # pragma: no cover
    assert parse_date(s1)-parse_date(s2) == timedelta(0), "%r != %r"%(s1, s2)

def _check_format(s, d): # pragma: no cover
    s2 = format_date(d)
    assert s == s2, "%r != %r"%(s, s2)

def test(): # pragma: no cover
    _check_fail("2007-03-23T05:27Z0500")
    _check_fail("200")
    _check_fail("garbage")
    _check_date("2007", datetime(2007, 1, 1, 0, 0, 0, 0, UTC), strict=False)
    _check_date("2007-03", datetime(2007, 3, 1, 0, 0, 0, 0, UTC), strict=False)
    _check_date("2007-03-23", datetime(2007, 3, 23, 0, 0, 0, 0, UTC), strict=False)
    _check_date("2007-3-23", datetime(2007, 3, 23, 0, 0, 0, 0, UTC), strict=False)
    _check_date("2007-3-3", datetime(2007, 3, 3, 0, 0, 0, 0, UTC), strict=False)
    _check_date("2007-03-23T05:27", datetime(2007, 3, 23, 5, 27, 0, 0, UTC), strict=True)
    _check_date("2007-03-23 05:27", datetime(2007, 3, 23, 5, 27, 0, 0, UTC), strict=True)
    _check_date("2007-03-23 05:27Z", datetime(2007, 3, 23, 5, 27, 0, 0, UTC), strict=True)
    _check_date("2007-03-23 05:27Z", datetime(2007, 3, 23, 5, 27, 0, 0, UTC), strict=True)
    _check_date("2007-03-23T05:27-0300",
                datetime(2007, 3, 23, 5, 27, 0, 0, TimeZone(-3*3600)), strict=True)
    _check_date("2007-03-23T05:27-300",
                datetime(2007, 3, 23, 5, 27, 0, 0, TimeZone(-3*3600)), strict=False)
    _check_date("2007-03-23T05:27-03",
                datetime(2007, 3, 23, 5, 27, 0, 0, TimeZone(-3*3600)), strict=False)
    _check_date("2007-03-23T05:27+2",
                datetime(2007, 3, 23, 5, 27, 0, 0, TimeZone(2*3600)), strict=False)
    _check_date("2007-03-23T05:27:23.023-0300",
                datetime(2007, 3, 23, 5, 27, 23, 23000, TimeZone(-3*3600)), strict=True)
    _check_date("2007-03-23T05:27:23.023-0300",
                datetime(2007, 3, 23, 5, 27, 23, 23000, TimeZone(-3*3600)), strict=True)
    _check_equal("2007-03-23T02:17-0330", "2007-03-23T05:47Z")

    # Check seconds since epoch calculations
    got = seconds_since_epoch("2007-01-25T12:30:00Z")
    expected = 1169728200
    #>>> iso8601.seconds_since_epoch("2007-01-25T12:00:00Z")
    #1169744400.0

    assert got == expected, "%s != %s"%(got, expected)
    got = seconds_since_epoch("2007-01-25T12:30:00-0100")
    assert got == expected+3600, "%s != %s"%(got, expected+3600)
    got = seconds_since_epoch("2007-01-25T12:30:00.1-0100")
    assert abs(got-(expected+3600.1)) < 1e-6, "%s != %s"%(got, expected+3600.1)

    # Determine local time offset
    hrs = abs(time.timezone)//3600
    mins = (abs(time.timezone)%3600)//60
    utcoffset = "%s%02d%02d"%('-' if time.timezone > 0 else '+', hrs, mins)

    # Check format from naive datetime object
    _check_format("2007-01-24T05:27:23"+utcoffset,
                  datetime(2007, 1, 24, 5, 27, 23, 16400))

    # Check formatting from struct_time object
    _check_format("2007-01-23T05:27:24"+utcoffset,
                  datetime(2007, 1, 23, 5, 27, 24).timetuple())

    # Check formatting from seconds since epoch
    got = format_date(seconds_since_epoch("2007-01-25T11:30:00-0100"))
    expected = format_date(datetime(2007, 1, 25, 12-hrs, 30-mins))
    assert got == expected, "%r != %r"%(got, expected)

    # Check formatting from zoned datetime objects
    expected = "2007-01-25T11:30:00-0100"
    got = format_date(parse_date(expected))
    assert got == expected, "%r != %r"%(got, expected)

    try: TimeZone(offset=1)
    except ValueError: pass
    else: raise Exception("exception not raised for TimeZone offset not minute")

    assert TimeZone(offset=3600).tzname(False) == "+0100"
    assert repr(TimeZone(offset=-3600)) == "<TimeZone '-0100'>"
    assert repr(TimeZone(name='-1', offset=-3600)) == "<TimeZone '-1'>"

if __name__ == "__main__": # pragma: no cover
    test()



from numpy import pi, arange, arctan2, sqrt, meshgrid, linspace, sin, cos, array, log, ones
from scipy import ndimage

def ConvertToCylindrical(array_in, x_min, x_max, y_min, y_max, theta_offset = 0.0, oversample_th = 2.0, oversample_r = 2.0):
    """Take an array in rectangular coordinates and transform into cylindrical
    (r,theta) coordinates.  Must specify axis limits (min and max) in real units
    of both axes.  As a side effect, this allows calculation of the center element ->
    (x0, y0) = (-x_min/x_stepsize, -y_min/y_stepsize)

    the spacing of the input mesh in the transformed variables is approximately:
    dr = (1/r) * (x dx + y dy)
    dtheta = 1/r^2 * sqrt(x^2 dy + y^2 dx)

    the output mesh is has a spacing of dr_min/oversample_r, and
    dtheta_min/oversample_th
    so that the output grid is bigger (larger array) than the input array, when
    oversample_r and oversample_th are greater than 1.

    At each pixel of the output array, a lookup pair of x, y is created
    that correspond to the r, th at that pixel.  The value of the input array
    at x, y is inserted into the output array (at r, th) via the scipy.ndimage function.

    Lookup pairs that are outside the x,y boundaries of the input array result in a zero
    being written at r,th (this occurs every time, as a rectangular r,th space maps onto
    a non-rectangular x,y space)

    A mask with the same dimensions as the output array is also generated, which has a
    value 1 everywhere the mapping was successful, and 0 where unsuccessful.

    Returns: data array in r,th space; mask; and boundaries of new r,th space
    """

    x_axis = linspace(x_min, x_max, array_in.shape[0])
    y_axis = linspace(y_min, y_max, array_in.shape[1])

    x_stepsize = float(x_max - x_min) / array_in.shape[0]
    y_stepsize = float(y_max - y_min) / array_in.shape[1]

    x,y = meshgrid(x_axis, y_axis)
    # meshgrid makes two new arrays with same dimensions as array_in,
    # but filled with x and y coordinates of each point instead of data

    r_in = sqrt(x**2 + y**2)
    theta_in = arctan2(y, x)
    # these are two more arrays with same dimensions as array_in,
    # but filled with the r, theta coordinates of each point of array_in

    dtheta_min = (1.0/r_in[r_in>0]**2 * sqrt(x[r_in>0]**2 * y_stepsize**2 + y[r_in>0]**2 * x_stepsize)).min() * 180.0/pi
    # dtheta is approximately 1/r^2 * sqrt(x^2 dy + y^2 dx)
    th_step = dtheta_min / oversample_th

    dr_min = (1.0/r_in[r_in>0] * ( abs(x[r_in>0] * x_stepsize) + abs(y[r_in>0] * y_stepsize) )).min()
    # dr is (1/r) * (x dx + y dy)
    r_step = dr_min / oversample_r

    th_out_axis = arange(theta_offset, theta_offset + 360.0 + th_step, th_step )
    r_out_axis = arange(r_in.min(), r_in.max(), r_step)

    th_out, r_out = meshgrid(th_out_axis, r_out_axis)

    x_lookup = ( r_out * cos(th_out*pi/180.0) - x_min ) / x_stepsize  # coordinates for looking up data in array_in
    y_lookup = ( r_out * sin(th_out*pi/180.0) - y_min ) / y_stepsize
    #return x_lookup, y_lookup, r_out, th_out

    data_mask = ones(r_out.shape, dtype=int)
    data_mask[x_lookup >= array_in.shape[0]] = 0
    data_mask[x_lookup < 0] = 0
    data_mask[y_lookup >= array_in.shape[1]] = 0
    data_mask[y_lookup < 0] = 0

    #x_lookup.clip(0)
    #y_lookup.clip(0)

    array_out = ndimage.map_coordinates(array_in, array([x_lookup, y_lookup]), order = 1)
    # ndimage.map_coordinates takes a source image, and a set of two arrays: the arrays hold lookup coordinates.
    # Each pixel of a new array (with dimensions equal to the lookup arrays) is set by looking at the pixel
    # in the source image with the coordinate pair specified.

    extent = [th_out.min(), th_out.max(), r_in.min(), r_in.max()]
    return array_out, data_mask, extent
    #self.qxpp.clip(0)
    #self.qzpp.clip(0)

if __name__ == '__main__':
    from pylab import imshow, show, xlabel, ylabel, colorbar, figure, title
    x, y = meshgrid(arange(-50., 150., 1), arange(-20., 30., 1))
    r = sqrt(x**2 + y**2)
    th = arctan2(y, x)
    d = sin(r*pi/25.0)**2
    #d = cos(th)**2 * 1.0/(r + 1.0)**2
    array_out, data_mask, extent = ConvertToCylindrical(d.T, -50, 150, -20, 30)
    # transposing the meshgrid because meshgrid(x,y) has shape (len(y), len(x))

    figure()
    imshow(array_out, aspect = 'auto', origin='lower', extent = extent)
    xlabel(r'$\theta ({}^{\circ})$', size='large')
    ylabel(r'$r$', size='large')
    title(r'$\sin^2(r)$ in cylindrical coords', size='large')
    colorbar()

    figure()
    imshow(d, origin='lower', extent=[-50,150,-20,30])
    xlabel(r'$x$', size='large')
    ylabel(r'$y$', size='large')
    title(r'$\sin^2(r)$ in rectangular coords', size='large')
    colorbar()

    a2, m2, e2 = ConvertToCylindrical(d.T, -50, 150, -20, 30, theta_offset=90.0)
    figure()
    imshow(a2, aspect = 'auto', origin='lower', extent=e2)
    xlabel(r'$\theta ({}^{\circ})$', size='large')
    ylabel(r'$r$', size='large')
    title(r'$\sin^2(r)$ in cylindrical coords with 90.0 deg offset', size='large')
    colorbar()
    show()





"""
Data loader for Bruker table-top X-ray source raw file format.
"""
import sys
import struct
import logging

import numpy

if sys.version_info[0] >= 3:
    def tostr(s):
        return s.decode('ascii')
else:
    def tostr(s):
        return s

MEAS_FLAG = {
    0: 'unmeasured',
    1: 'measured',
    2: 'active',
    3: 'aborted',
    4: 'interrupted',
    10: '.SIM file K.S.',
    }
SCAN_TYPE = {
    0: 'locked coupled',    # twotheta, with theta = twotheta/2
    1: 'unlocked coupled',  # twotheta, with theta = twotheta/2 + initial offset
    2: 'detector scan',     # twotheta
    3: 'rocking curve',     # theta
    4: 'chi scan',          # sample tilt
    5: 'phi scan',          # sample rotation
    6: 'x scan',            # sample translation x
    7: 'y scan',            # sample translation y
    8: 'z scan',            # sample translation z
    9: 'aux1 scan',
    10: 'aux2 scan',
    11: 'aux3 scan',
    12: 'psi scan',
    13: 'hkl scan',
    14: 'rec. space scan',
    20: 'unlocked coupled HR XRD',
    129: 'PSD fixed scan',
    130: 'PSD fast scan',
    }

VARYING_BIT = (
    'two_theta',
    'theta',
    'chi',
    'phi',
    'x',
    'y',
    'z',
    'aux1',
    'aux2',
    'aux3',
    'time',
    'temp',
    )

def _make_struct(fields, data, offset=0):
    names = list(zip(*fields))[0]
    types = "<"+"".join(dtype for _, dtype in fields)
    values = struct.unpack_from(types, data, offset=offset)
    new_values = []
    for (_, t), v in zip(fields, values):
        if t[-1] == 's':
            try:
                output = tostr(v).strip('\0')
            except UnicodeDecodeError:
                output = "ERROR: could not decode"
            new_values.append(output)
        else:
            new_values.append(v)
    offset += struct.calcsize(types)
    return dict(zip(names, new_values)), offset

RAW_HEADER = (
    ('raw_id', '8s'),              # 0
    ('meas_flag', 'i'),            # 8 0: unmeasured, 1: measured, 2: active, 3: aborted, 4: interrupted
    ('no_of_tot_meas_ranges', 'i'), # 12
    ('date', '10s'),               # 16 mm/dd/yy
    ('time', '10s'),               # 26 hh:mm:ss
    ('user', '72s'),               # 36
    ('site', '218s'),              # 108
    ('samplename', '60s'),         # 326
    ('comment', '160s'),           # 386
    ('res0', '2s'),                # 546 2 byte alignment
    ('goniom_model', 'i'),         # 548
    ('goniom_stage', 'i'),         # 552
    ('sample_changer', 'i'),       # 556
    ('goniom_ctrl', 'i'),          # 560
    ('goniom_radius', 'f'),        # 564 diameter of the goniom (mm)
    ('fixed_inc_divsli', 'f'),     # 568 diverg.slit fix incidence
    ('fixed_inc_samplesli', 'f'),  # 572 sample slit fix inc.
    ('fixed_inc_soller', 'i'),     # 576 sollerslit 0: None, 1: 2, 2: 4 deg
    ('fixed_inc_monochromator', 'i'), # 580
    ('fixed_dif_antisli', 'f'),    # 584 anti slit diffracted site
    ('fixed_dif_detsli', 'f'),     # 588 detecgtor slit diffr. site
    ('fixed_dif_2ndsoller', 'i'),  # 592 soller slit 0: None, 1: 2, 2: 4 deg
    ('fixed_dif_thinfilm', 'i'),   # 596 thin film 0: None, 1: 0.4, 2: 0.15, 3: 0.08 deg
    ('fixed_dif_betafilter', 'i'), # 600
    ('fixed_dif_analyzer', 'i'),   # 604
    ('anode', '4s'),               # 608 anode material
    #('res1', '4s'),               # 612 8 byte alignment
    ('actual_QUANT_offset', 'f'),  # 612 the actual offset used by QUANT
    ('alpha_average', 'd'),        # 624 (alpha_1+alpha_2)/2 Ang
    ('alpha_1', 'd'),              # 632 alpha1 wavelength Ang
    ('alpha_2', 'd'),              # 640 alpha2 wavelength Ang
    ('beta', 'd'),                 # 648
    ('alpha_21', 'd'),             # 652 alpha_2/alpha_1
    ('wave_unit', '4s'),           # 656 A or nm; not really used
    ('beta_rel_int', 'f'),         # 660 relative int.of beta
    ('total_sample_run_time', 'f'),# 664
    #('reserved', '36s'),          # 668
    ('reserved', '32s'),           # 668
    ('PSDopening', 'f'),           # 700 angular opening in deg
    ('no_of_tot_meas_ranges_in_dq1', 'i'), # 704
    ('reserved_1', '1s'),          # 708
    ('further_dql_reading', '1s'), # 709
    ('cQuantSequence', '1s'),      # 710
    ('HWindicator', '1s'),         # 711
    )

RAW_RANGE_HEADER = (
    ('length_of_RAW_RANGE_HEADER', 'i'),# 0 header only, not optional parameters
    ('no_of_measured_data', 'i'),  # 4
    ('theta_start', 'd'),          # 8
    ('two_theta_start', 'd'),      # 16
    ('chi_start', 'd'),            # 24
    ('phi_start', 'd'),            # 32
    ('x_start', 'd'),              # 40
    ('y_start', 'd'),              # 48
    ('z_start', 'd'),              # 56
    ('divslit_start', 'd'),        # 64 Div.slit start if divslit_code==fix
    ('divslit_code', '6s'),        # 72 V20, V6, V4, Vxxx, unkn, or fix
    ('res2', '2s'),                # 78
    ('antislit_start', 'd'),       # 80 Anti.slit start if antislit_code==fix
    ('antislit_code', '6s'),       # 88 V20, V6, V4, Vxxx, unkn, or fix
    ('res3', '2s'),                # 94
    ('detector_1', 'i'),           # 96
    ('det_HV_1', 'f'),             # 100 high voltage
    ('det_AG_1', 'f'),             # 104 amplifier gain
    ('det_LL_1', 'f'),             # 108 lower level
    ('det_UL_1', 'f'),             # 112 upper level
    ('detector_2', 'i'),           # 116 measuring channel
    ('res4', '8s'),                # 120
    ('det_LL_2', 'f'),             # 128 Lower level in % of full scale
    ('det_UL_2', 'f'),             # 132 Upper level in % of full scale
    ('detslit_code', '5s'),        # 136 in, out, unkn
    ('res5', '3s'),                # 141
    ('aux1_start', 'd'),           # 144 start of auxil. axis 1 (up to now 0.0 )
    ('aux2_start', 'd'),           # 152
    ('aux3_start', 'd'),           # 160
    ('scan_mode', 'i'),            # 168 0: step (SS), 1: continuous (SC)
    ('res6', '4s'),                # 172
    ('increment_1', 'd'),          # 176 increment of the scan
    ('increment_2', 'd'),          # 184 unused, so 0.0
    ('step_time', 'f'),            # 192
    ('scan_type', 'i'),            # 196 SCAN_TYPE
    ('meas_delay_time', 'f'),      # 200 delay time before measurement
    ('range_sample_started', 'f'), # 204 the time the range was started relative to the sample
    ('rot_speed', 'f'),            # 208
    ('temperature', 'f'),          # 212 temperature in K
    ('heating_cooling_rate', 'f'), # 216
    ('temp_delay_time', 'f'),      # 220
    ('generator_voltage', 'i'),    # 224 kV
    ('generator_current', 'i'),    # 228 mA
    ('display_plane_number', 'i'), # 232 needed for 3d display
    ('res7', '4s'),                # 236
    ('act_used_lambda', 'd'),      # 240 actually selected wavelength
    ('varying_parameters', 'i'),   # 248 VARYING_BIT
    ('data_record_length', 'i'),   # 252 4 + 8 x num bits in varying_parameters
    ('total_size_of_extra_records', 'i'), # 256
    ('smoothing_width', 'f'),      # 260
    ('sim_meas_cond', 'i'),        # 264
    ('res8', '4s'),                # 268
    ('increment_3', 'd'),          # 272
    ('reserved', '24s'),           # 280
    )

EXTRA_RECORD = {
    100: ( # OSC
        ('osci_circle', 'i'),
        ('res9', '4s'),
        ('osci_amplitude', 'd'),
        ('osci_speed', 'f'),
        ('reserved', '12s'),
        ),
    110: ( # PSD
        ('act_two_theta', 'd'),
        ('first_unused_channel', 'i'),
        ('reserved', '20s'),
        ),
    120: ( # OQM
        ('mode', 'i'),
        ('two_theta_actual', 'f'),
        ('time_actual', 'f'),
        ('counts', 'f'),
        ('compound', '32s'),
        ('peak_id', 'i'),
        ('reserved', '60s'),
        ),
    150: ( # ORM
        ('first_two_theta', 'f'),
        ('last_two_theta', 'f'),
        ('reserved', '16s'),
        ),
    190: ( # OLC
        ('two_theta_offset', 'f'),
        ('int_offset', 'f'),
        ('reserved', '16s'),
        ),
    200: ( # AD
        ('act_two_theta', 'd'),
        ('tthbeg', 'f'),
        ('tthend', 'f'),
        ('chibeg', 'f'),
        ('chiend', 'f'),
        ('normal', 'i'),
        ('program', '20s'),
        ('reserved', '16s'),
        ),
    300: ( # HRXRD
        ('szSubstrate', '40s'),
        ('sz_hkl', '9s'),
        ('sz_mno', '9s'),
        ('sz_pqr', '9s'),
        ('szDiffractionSetup', '1s'),
        ('fAnalyzerOffs', 'f'),
        ('iIncidenceFlag', 'i'),
        ('iAlignmentSet', 'i'),
        ('fLambdaTheory', 'd'),
        ('fLambdaDetermined', 'd'),
        ('iHRXRDScanType', 'i'),
        ('iMonochrAnalysComb', 'i'),
        ('fSim_FF', 'd'),
        ('fCoupleFactor', 'd'),
        ('fThetaBragg', 'd'),
        ('fTauBragg', 'd'),
        ('fTauMis', 'd'),
        ('fLattice', 'd'),
        ('fPsi', 'd'),
        ('fScanRelStart', 'd'),
        ('fScanRelStop', 'd'),
        ('fOmegaMiddle', 'd'),
        ('f2ThetaMiddle', 'd'),
        ('fThetaMiddle', 'd'),
        ('fPhiMiddle', 'd'),
        ('fChiMiddle', 'd'),
        ('fXMiddle', 'd'),
        ('fYMiddle', 'd'),
        ('fZMiddle', 'd'),
        ('fQParallelMiddle', 'd'),
        ('fQVerticalMiddle', 'd'),
        ('fHeidenhainPeak', 'd'),
        ('fXOriginal', 'd'),
        ('fYOriginal', 'd'),
        ('fZOriginal', 'd'),
        ('fOmegaOrigianl', 'd'),
        ('fOmege1Original', 'd'),
        ('fOmega2Original', 'd'),
        ('iSimLength', 'i'),
        # needs simdat, simusubstrate, layer block header
        ),
    10140: ( # OCM
        ('comment', '0s'),
        # needs comment, which is total length of the record
        ),
    10250: ( # REFSIM
        ('ABS_FLAG', '1s'),
        ('SIZ_FLAG', '1s'),
        ('BEAM_FLAG', '1s'),
        ('SWIDTH_FLAG', '1s'),
        ('SDIST_FLAG', '1s'),
        ('res1', '3s'),
        ('SamSize', 'f'),
        ('BeamSize', 'f'),
        ('slitwidth', 'f'),
        ('slitdist', 'f'),
        ('SimType', '1s'),
        ('res2', '3s'),
        ('abs_fac', 'f'),
        ('layers', 'h'),
        ('multis', 'h'),
        ('variable_data1', 'i'),
        ('variable_data2', 'i'),
        # needs rlayer, mulit
        ),
    }


def load(filename):
    """
    Load Bruker X-ray data from *filename*.
    """
    with open(filename, 'rb') as f:
        data = f.read()
    if data[:7] != b"RAW1.01":
        raise ValueError("Could not load %r: not a Bruker XRD RAW file"%
                         filename)
    return loads(data)

def loads(data):
    """
    Load Bruker X-ray data from the byte stream contained in *data*.
    """
    if data[:7] != b"RAW1.01":
        raise ValueError("not a Bruker XRD RAW file")

    # process the main header
    offset = 0
    header, offset = _make_struct(RAW_HEADER, data, offset)
    #assert offset == 712


    # for each range, read in the range
    ranges = []
    for _ in range(header['no_of_tot_meas_ranges']):
        # range starts with the range header
        rheader, offset = _make_struct(RAW_RANGE_HEADER, data, offset)
        extra_length = rheader['total_size_of_extra_records']

        # can have multiple extra records, so append them as they appear
        rheader['extra'] = []
        while extra_length > 0:
            # find the extra record type, and use it to parse the extra data
            rectype, reclen = struct.unpack_from('ii', data, offset=offset)
            if rectype not in EXTRA_RECORD:
                raise ValueError('unknown measurement type %d'%rectype)
            rextra, _ = _make_struct(EXTRA_RECORD[rectype], data, offset+8)
            # note: some extra records (e.g., HRXRD simulations) have variable
            # data stored after the record.  We're skipping this for now.

            # plug record type and length into each record
            rextra['record_type'] = rectype
            rextra['record_length'] = reclen

            # save the extra in the range header
            rheader['extra'].append(rextra)

            # move to the next extra header
            offset += reclen
            extra_length -= reclen

        # parse the data block
        nrow = rheader['no_of_measured_data']
        ncol = (rheader['data_record_length']-4)//8
        types = ('f'+'d'*ncol)*nrow
        values = struct.unpack_from(types, data, offset=offset)
        offset += nrow * (4 + 8*ncol)
        columns = numpy.array(values, 'd').reshape((ncol+1, nrow))

        # figure out which columns are in use
        colnames = [n for i, n in enumerate(VARYING_BIT)
                    if (2**i)&rheader['varying_parameters']]
        if len(colnames) != ncol:
            raise ValueError('varying_parameters and data_record_length are inconsistent')
        values = {'count': columns[0]}
        values.update((n, v) for n, v in zip(colnames, columns[1:]))

        rheader['values'] = values
        ranges.append(rheader)

    header['data'] = ranges
    return header

if __name__ == "__main__":
    import sys, pprint
    pprint.pprint(load(sys.argv[1]))



from functools import wraps

from .FilterableMetaArray import FilterableMetaArray as MetaArray
from .magik_filters import Filter2D, Algebra, updateCreationStory, autoApplyToList

class AsterixPixelsToTwotheta(Filter2D):
    """ input array has pixels axis, convert to
    two-theta based on distance from sample to detector and
    center pixel when two-theta motor is set to zero """

    @autoApplyToList
    @updateCreationStory
    def apply(self, data, qzero_pixel = 145., twotheta_offset=0.0, pw_over_d=0.0003411385649):
        """ pw_over_d is pixel width divided by distance (sample to detector) """
        new_info = data.infoCopy()
        # find pixels axis, and replace with two-theta
        # assuming there are only 2 dimensions of data, looking only at indices 0, 1
        xpixel_axis = next((i for i in range(len(new_info)) if new_info[i]['name'] == 'xpixel'), None)
        if xpixel_axis < 0:
            print("error: no xpixel axis in this dataset")
            return

        new_info[xpixel_axis]['name'] = 'twotheta'
        twotheta_motor = 0.0
        if data._info[-1].has_key('state'):
            twotheta_motor = float(data._info[-1]['state']['A[0]'])
        pixels = data.axisValues('xpixel')
        twotheta = arctan2((pixels - qzero_pixel) * pw_over_d, 1.0) * 180./pi + twotheta_offset + twotheta_motor
        new_info[xpixel_axis]['values'] = twotheta
        new_data = MetaArray(data.view(ndarray).copy(), info=new_info)

        return new_data

class AsterixTOFToWavelength(Filter2D):
    """ input array has TOF axis, convert to wavelength
    based on calibration (depends on distance from source to detector) """

    @autoApplyToList
    @updateCreationStory
    def apply(self, data, wl_over_tof=1.9050372144288577e-5):
        """ wl_over_tof is wavelength divided by time-of-flight """
        new_info = data.infoCopy()
        # find pixels axis, and replace with two-theta
        # assuming there are only 2 dimensions of data, looking only at indices 0, 1
        tof_axis = next((i for i in range(len(new_info)) if new_info[i]['name'] == 'tof'), None)
        if tof_axis < 0:
            print("error: no tof axis in this dataset")
            return

        new_info[tof_axis]['name'] = 'wavelength'
        tof = data.axisValues('tof')
        wavelength = (tof * wl_over_tof)
        # shift by half-bin width to align to center of tof bins!
        wavelength += (tof[1] - tof[0])/2.0 * wl_over_tof
        # (bins appear to be centered)
        new_info[tof_axis]['values'] = wavelength
        new_data = MetaArray(data.view(ndarray).copy(), info=new_info)

        return new_data

class AsterixShiftData(Filter2D):

    @autoApplyToList
    @updateCreationStory
    def apply(self, data, edge_bin = 180, axis=0):
        """ Shift 2D dataset along axis 0, also shifting the axisValues
        along that edge (assuming linear behaviour)
        This is useful for time-of-flight data where the low-t data is empty due
        to spectrum shape, and can be interpreted as the high-t data from the
        previous pulse."""
        #axis = 0
        axis = int(axis)
        if axis > 1:
            axis = 1
        if axis < 0:
            axis = 0
        new_info = data.infoCopy()
        old_axis_values = new_info[axis]['values']
        src_data = data.view(ndarray).copy()

        shifted_data = empty(src_data.shape)

        #start with empty slices over first two axes
        output_slice = [slice(None, None)] * len(src_data.shape)
        input_slice = [slice(None, None)] * len(src_data.shape)

        # move data from edge_bin to end to beginning:
        input_slice[axis] = slice(edge_bin, None)
        output_slice[axis] = slice(None, -edge_bin)
        shifted_data[output_slice] = src_data[input_slice]
        # move data from beginning to edge bin to the end of the new dataset
        input_slice[axis] = slice(None, edge_bin)
        output_slice[axis] = slice(-edge_bin, None)
        shifted_data[output_slice] = src_data[input_slice]

        shifted_axis = zeros(src_data.shape[axis])
        dx = old_axis_values[1] - old_axis_values[0]
        shifted_axis[:-edge_bin] = old_axis_values[edge_bin:]
        shifted_axis[-edge_bin:] = old_axis_values[:edge_bin] + (old_axis_values[-1] - old_axis_values[0]) + dx

        new_info[axis]['values'] = shifted_axis
        new_data = MetaArray(shifted_data, info=new_info)
        return new_data

class AsterixCorrectSpectrum(Filter2D):

    def apply(self, data, spectrum):
        ##polarizations = ["down_down", "down_up", "up_down", "up_up"]
        # inherit polarizations list from Filter2D:
        passthrough_cols = ["counts_%s" % (pol,) for pol in self.polarizations]
        passthrough_cols.extend(["pixels", "count_time"])
        #expressions = [{"name":col, "expression":"data1_%s" % (col,)} for col in passthrough_cols]
        expressions = []
        spectrum_cols = [col['name'] for col in spectrum._info[-2]['cols']]
        for pol in self.polarizations:
            if ("spectrum_%s" % (pol,) in spectrum_cols):
                spec_id = "data2_spectrum_%s" % (pol,)
            else:
                spec_id = "data2_spectrum"
            expressions.append({"name":"monitor_%s" % (pol,), "expression":"data1_monitor_%s * %s[:,newaxis]" % (pol,spec_id)})
        result = Algebra().apply(data, spectrum, expressions, passthrough_cols)
        return result

class TwothetaLambdaToQxQz(Filter2D):
    """ Figures out the Qx, Qz values of each datapoint
    and throws them in the correct bin.  If no grid is specified,
    one is created that covers the whole range of Q in the dataset

    If autofill_gaps is True, does reverse lookup to plug holes
    in the output (but pixel count is still zero for these bins)
    """

    default_qxqz_gridvals = (-0.003, 0.003, 201, 0, 0.1, 201)

    def getQxQz (self, theta, twotheta, wavelength = 5.0):
        qLength = 2.0 * pi / wavelength
        tilt = theta - ( twotheta / 2.0 )
        dq = 2.0 * qLength * sin( ( pi / 180.0 ) * ( twotheta / 2.0 ) )
        qxOut = dq * sin( pi * tilt / 180.0 )
        qzOut = dq * cos( pi * tilt / 180.0 )
        return [qxOut, qzOut]

    @autoApplyToList
    #@updateCreationStory
    def apply(self, data, theta=None, qxmin=None, qxmax=None, qxbins=None, qzmin=None, qzmax=None, qzbins=None):
        info = [{"name": "qx", "units": "inv. Angstroms", "values": linspace(qxmin, qxmax, qxbins) },
                {"name": "qz", "units": "inv. Angstroms", "values": linspace(qzmin, qzmax, qzbins) },]
        old_info = data.infoCopy()
        info.append(old_info[2]) # column information!
        info.append(old_info[3]) # creation story!
        output_grid = MetaArray(zeros((qxbins, qzbins, data.shape[-1])), info=info)


        #if output_grid == None:
        #    output_grid = EmptyQxQzGrid(*self.default_qxqz_gridvals)
        #else:
        #    output_grid = deepcopy(output_grid)

        if (theta == "") or (theta == None):
            if 'state' in data._info[-1]:
                theta = float(data._info[-1]['state']['A[1]'])
                print('theta:', theta)
            else:
                print("can't run without theta!")
                return

        wl_array = data.axisValues('wavelength').copy()
        wl_array.shape = wl_array.shape + (1,)
        twotheta_array = data.axisValues('twotheta').copy()
        twotheta_array.shape = (1,) + twotheta_array.shape
        qxOut, qzOut = self.getQxQz(theta, twotheta_array, wl_array)

        # getting values from output grid:
        outgrid_info = output_grid.infoCopy()
        numcols = len(outgrid_info[2]['cols'])
        qx_array = output_grid.axisValues('qx')
        dqx = qx_array[1] - qx_array[0]
        qz_array = output_grid.axisValues('qz')
        dqz = qz_array[1] - qz_array[0]
        framed_array = zeros((qz_array.shape[0]+2, qx_array.shape[0]+2, numcols))
        target_qx = ((qxOut - qx_array[0])/dqx + 1).astype(int)
        #return target_qx, qxOut
        target_qz = ((qzOut - qz_array[0])/dqz + 1).astype(int)
        target_mask = (target_qx >= 0) * (target_qx < qx_array.shape[0])
        target_mask *= (target_qz >= 0) * (target_qz < qz_array.shape[0])
        target_qx_list = target_qx[target_mask]
        target_qz_list = target_qz[target_mask]
        #target_qx = target_qx.clip(0, qx_array.shape[0]+1)
        #target_qz = target_qz.clip(0, qz_array.shape[0]+1)

        for i, col in enumerate(outgrid_info[2]['cols']):
            values_to_bin = data[:,:,col['name']][target_mask]
            outshape = (output_grid.shape[0], output_grid.shape[1])
            hist2d, xedges, yedges = histogram2d(target_qx_list,target_qz_list, bins = (outshape[0],outshape[1]), range=((0,outshape[0]),(0,outshape[1])), weights=values_to_bin)
            output_grid[:,:,col['name']] += hist2d
            #framed_array[target_qz_list, target_qx_list, i] = data[:,:,col['name']][target_mask]

        #trimmed_array = framed_array[1:-1, 1:-1]
        #output_grid[:,:] = trimmed_array

        creation_story = data._info[-1]['CreationStory']
        new_creation_story = creation_story + ".filter('{0}', {1})".format(self.__class__.__name__, output_grid._info[-1]['CreationStory'])
        #print(new_creation_story)
        output_grid._info[-1] = data._info[-1].copy()
        output_grid._info[-1]['CreationStory'] = new_creation_story
        return output_grid



"""
Dataflow architecture.

The dataflow architecture provides reduction and analysis routines for 
a variety neutron scattering instruments.

Operations are organized by instruments, each with its own data formats
and transformations.  These are defined as :class:`.core.Instrument`,
:class:`.core.Datatype` and :class:`.core.Module` respectively.  Modules
have an icon to display on the screen, fields for defining the parameters
for the computation, terminals for defining the data types that can flow
in and out of the module.

Each instrument will have a standard set of templates (defined as
:class:`.core.Template`) which control the order of operations
in the data flow.  A template is a list of modules, the connections
between them, and a configuration set which gives initial values to each
of the fields.   When templates are instantiated by the user, and values
are given for all the configuration parameters, the template is
evaluated by :function:`.calc.process_template`.  The *proccess_template*
function takes care of the order of evaluation, computing only what is needed
for the desired output.  The evaluation produces data objects of the
instrument-defined data types which are shared between components or sent to
the client for display.

Instruments, Modules and Datatypes are registered with :module:`core`.  They
are then available to our web-based reduction software, which allows users
to define and fill templates and display the raw and reduced data.
"""

__version__ = "0.2"



from reductus.dataflow import core as df
from reductus.dataflow.automod import make_modules

from . import steps
from . import templates
from .vsansdata import RawVSANSData, VSansDataRealSpace, VSansDataQSpace, Parameters, VSans1dData #, Metadata

INSTRUMENT = "ncnr.vsans"

def define_instrument():
    # Define modules
    modules = make_modules(steps.ALL_ACTIONS, prefix=INSTRUMENT+'.')

    # Define data types
    vsans_raw = df.DataType(INSTRUMENT+".raw", RawVSANSData)
    vsans_realspace = df.DataType(INSTRUMENT+".realspace", VSansDataRealSpace)
    vsans_qspace = df.DataType(INSTRUMENT+".qspace", VSansDataQSpace)
    vsans_1d = df.DataType(INSTRUMENT+'.v1d', VSans1dData)
    #vsans_metadata = df.DataType(INSTRUMENT+".metadata", Metadata)
    params = df.DataType(INSTRUMENT+".params", Parameters)

    # Define instrument
    vsans = df.Instrument(
        id=INSTRUMENT,
        name='NCNR Very Small Angle Neutron Scattering (VSANS) instrument',
        menu=[('steps', modules)],
        datatypes=[vsans_raw, vsans_realspace, vsans_qspace, params, vsans_1d],
        template_defs=df.load_templates(templates),
        )

    # Register instrument
    df.register_instrument(vsans)
    return vsans



"""
Get commit id from the git repo.

Drop the file rev.py into a top level directory PACKAGE_NAME of your
application, right below the root of the repository.  Set the PACKAGE_NAME
variable in this file to the package name.  For example::

    PACKAGE_NAME = "reductus"

From within your application you can then do::

    from . import rev

    rev.print_revision()  # print the repo version
    commit = rev.revision_info()  # return commit id

If you use a more complicated source tree then you will need to replace
repo_path() with a function that returns the path to the repo root before
requesting the revision information.

On "pip install" the repo root directory will not be available. In this
case the code looks for PACKAGE_NAME/git_revision, which you need to
install into site-pacakges along with your other sources.

The simplest way to create PACKAGE_NAME/git_revision is to run rev
from setup.py::

    import sys
    import os

    # Create the resource file git_revision.
    if os.system(f'"{sys.executable}" PACKAGE_NAME/rev.py') != 0:
        print("setup.py failed to build PACKAGE_NAME/git_revision", file=sys.stderr)
        sys.exit(1)

    ...

    # Include git revision in the package data, eitherj by adding
    # "include PACKAGE_NAME/git_revision" to MANIFEST.in, or by
    # adding the following to setup.py:
    #package_data = {"PACKAGE_NAME": ["git_revision"]}
    setup(
        ...
        #package_data=package_data,
        include_package_data=True,
        ...
    )

Add the following to .gitignore, substituting your package name::

    /PACKAGE_NAME/git_revision

"""
from pathlib import Path
from warnings import warn

def repo_path():
    return Path(__file__).parent.parent.absolute()

def print_revision():
    """Print the git revision"""
    revision = revision_info()
    print("git revision", revision)

def store_revision():
    """
    Call from setup.py to save the git revision to the distribution.

    See :mod:`rev` for details.
    """
    commit = revision_info()
    if commit is None:
        commit = "unknown"
    path = Path(__file__).absolute().parent / RESOURCE_NAME
    with path.open("w") as fd:
        fd.write(commit + "\n")


PACKAGE_NAME = "reductus"
RESOURCE_NAME = "git_revision"
_REVISION_INFO = None # cached value of git revision
def revision_info():
    """
    Get the git hash and mtime of the repository, or the installed files.
    """
    # TODO: test with "pip install -e ." for developer mode
    global _REVISION_INFO

    if _REVISION_INFO is None:
        _REVISION_INFO = git_rev(repo_path())

    if _REVISION_INFO is None:
        try:
            from importlib import resources
        except ImportError: # CRUFT: pre-3.7 requires importlib_resources
            import importlib_resources as resources
        try:
            revdata = resources.read_text(PACKAGE_NAME, RESOURCE_NAME)
            commit = revdata.strip()
            _REVISION_INFO = commit
        except Exception:
            _REVISION_INFO = "unknown"

    return _REVISION_INFO

def git_rev(repo):
    """
    Get the git revision for the repo in the path *repo*.

    Returns the commit id of the current head.

    Note: this function parses the files in the git repository directory
    without using the git application.  It may break if the structure of
    the git repository changes.  It only reads files, so it should not do
    any damage to the repository in the process.
    """
    # Based on stackoverflow am9417
    # https://stackoverflow.com/questions/14989858/get-the-current-git-hash-in-a-python-script/59950703#59950703
    if repo is None:
        return None

    git_root = Path(repo) / ".git"
    git_head = git_root / "HEAD"
    if not git_head.exists():
        return None

    # Read .git/HEAD file
    with git_head.open("r") as fd:
        head_ref = fd.read()

    # Find head file .git/HEAD (e.g. ref: ref/heads/master => .git/ref/heads/master)
    if not head_ref.startswith("ref: "):
        warn("expected 'ref: path/to/head' in {git_head}".format(git_head=git_head))
        return None
    head_ref = head_ref[5:].strip()

    # Read commit id from head file
    head_path = git_root.joinpath(*head_ref.split("/"))
    if not head_path.exists():
        warn("path {head_path} referenced from {git_head} does not exist".format(head_path=head_path, git_head=git_head))
        return None

    with head_path.open("r") as fd:
        commit = fd.read().strip()

    return commit

def main():
    """
    When run as a python script create git_revision in the current directory.
    """
    print_revision()
    store_revision()

if __name__ == "__main__":
    main()



from __future__ import print_function

import sys
from collections import OrderedDict

import numpy as np

if sys.version_info[0] > 2:
    def tostr(s):
        return s.decode('utf8')
    def decode(s, encoding='utf8'):
        return s.decode(encoding)
    STR_ENCODING = 'utf8'
else:
    def tostr(s):
        return s
    def decode(s, encoding='utf8'):
        return unicode(s, encoding)
    STR_ENCODING = None

DATA_TYPES = {
    1: "scalar",
    2: "matrix",
    3: "complex scalar",
    4: "complex matrix",
    5: "old_string",
    6: "range",
    7: "string",
}

TYPE_CODES = {
    0: "u1",
    1: "u2",
    2: "u4",
    3: "i1",
    4: "i2",
    5: "i4",
    6: "f4",
    7: "f8",
    8: "u8",
    9: "i8",
}
DTYPES = {k: np.dtype(v) for k, v in TYPE_CODES.items()}

def loadoct(fd, encoding=STR_ENCODING):
    """
    Read an octave binary file from the file handle fd, returning
    an array of structures.  If encoding is not None then convert
    strings from bytes to unicode.  Default is STR_ENCODING, which
    is utf8 for python 3 and None for python 2, yielding arrays
    of type str in each dialect.
    """
    magic = fd.read(10)
    assert(magic == b"Octave-1-L" or magic == b"Octave-1-B")
    endian = "<" if magic[-1:] == b"L" else ">"
    # Float type is 0: IEEE-LE, 1: IEEE-BE, 2: VAX-D, 3: VAX-G, 4: Cray
    # Not used since Octave assumes IEEE format floats.
    _float_format = fd.read(1)
    len_dtype = np.dtype(endian + "i4")
    def read_len():
        len_bytes = fd.read(4)
        if not len_bytes:
            return None
        return np.frombuffer(len_bytes, len_dtype)[0]
    table = OrderedDict()
    while True:
        name_length = read_len()
        if name_length is None:  # EOF
            break
        name = tostr(fd.read(name_length))
        doc_length = read_len()
        doc = tostr(fd.read(doc_length)) if doc_length else ''
        is_global = bool(ord(fd.read(1)))
        data_type = ord(fd.read(1))
        if data_type == 255:
            type_str = tostr(fd.read(read_len()))
        else:
            type_str = DATA_TYPES[data_type]
        #print("reading", name, type_str)
        if type_str.endswith("scalar"):
            if type_str == "scalar":
                dtype = DTYPES[ord(fd.read(1))]
            elif type_str == "complex scalar":
                _ = fd.read(1)
                dtype = np.dtype('complex128')
            elif type_str == "float complex scalar":
                _ = fd.read(1)
                dtype = np.dtype('complex64')
            else:
                dtype = np.dtype(type_str[:-7])
            dtype = dtype.newbyteorder(endian)
            data = np.frombuffer(fd.read(dtype.itemsize), dtype)
            table[name] = data[0]
        elif type_str.endswith("matrix"):
            ndims = read_len()
            if ndims < 0:
                ndims = -ndims
                dims = np.frombuffer(fd.read(4*ndims), len_dtype)
            else:
                dims = (ndims, read_len())
            count = np.prod(dims)
            if type_str == "matrix":
                dtype = DTYPES[ord(fd.read(1))]
            elif type_str == "complex matrix":
                _ = fd.read(1)
                dtype = np.dtype('complex128')
            elif type_str == "float complex matrix":
                _ = fd.read(1)
                dtype = np.dtype('complex64')
            else:
                dtype = np.dtype(type_str[:-7])
            dtype = dtype.newbyteorder(endian)
            data = np.frombuffer(fd.read(count*dtype.itemsize), dtype)
            # Note: Use data.copy() to make a modifiable array.
            table[name] = data.reshape(dims, order='F')
        elif type_str == "old_string":
            data = fd.read(read_len())
            if encoding is not None:
                data = decode(data, encoding)
            table[name] = data
        elif type_str in ("string", "sq_string"):
            nrows = read_len()
            if nrows < 0:
                ndims = -nrows
                dims = np.frombuffer(fd.read(4*ndims), len_dtype)
                count = np.prod(dims)
                fortran_order = np.frombuffer(fd.read(count), dtype='uint8')
                c_order = np.ascontiguousarray(fortran_order.reshape(dims, order='F'))
                data = c_order.view(dtype='|S'+str(dims[-1]))
                if encoding is not None:
                    data = np.array([decode(s, encoding) for s in data.flat])
                table[name] = data.reshape(dims[:-1])
            else:
                data = [fd.read(read_len()) for _ in range(nrows)]
                if encoding is not None:
                    data = [decode(s, encoding) for s in data]
                table[name] = np.array(data)

        else:
            raise NotImplementedError("unknown octave type "+type_str)
        #print("read %s:%s"%(name, type_str), table[name])
    return table

read_octave_binary = loadoct  # CRUFT: deprecated name

def _dump(filename, encoding=STR_ENCODING):
    import gzip

    if filename.endswith('.gz'):
        with gzip.open(filename, 'rb') as fd:
            table = loadoct(fd, encoding)
    else:
        with open(filename, 'rb') as fd:
            table = loadoct(fd, encoding)
    for k, v in table.items():
        print(k, v)

if __name__ == "__main__":
    #_dump(sys.argv[1], encoding='utf8')  # unicode
    #_dump(sys.argv[1], encoding=None)  # bytes
    _dump(sys.argv[1])  # str, encoding=STR_ENCODING




#############################################################
# rename or copy this file to config.py if you make changes #
#############################################################

# change this to your fully-qualified domain name to run a
# remote server.  The default value of localhost will
# only allow connections from the same computer.
#
# for remote (cloud) deployments, it is advised to remove
# the "local" data_sources item below, and to serve static
# files using a standard webserver
#
# if use_redis is False, server will use in-memory cache.

# TODO: Convert this to JSON file in web-accesible ('static')
# directory.

config = {
    # ssl_args for https serving the rpc
    # ssl_args = {"keyfile": None, "certfile": None}

    # Cache engines are diskcache, redis, or memory if not specified
    "cache": {
        "engine": "diskcache", 
        "params": {"size_limit": int(4*2**30)}
    },
    "data_sources": [
        {
            "name": "local",
            "url": "file:///",
            "start_path": "",
        },
        {
            "name": "ncnr",
            "url": "https://ncnr.nist.gov/pub/",
            "start_path": "ncnrdata",
            "file_helper_url": "https://ncnr.nist.gov/ncnrdata/listftpfiles_json.php",
        },
        {
            "name": "charlotte",
            "url": "http://charlotte.ncnr.nist.gov/pub",
            "start_path": "",
            "file_helper_url": "http://charlotte.ncnr.nist.gov/ncnrdata/listftpfiles_json.php",
        },
        # set start_path for local files to usr/local/nice/server_data/experiments
        # for instrument installs
        {
            "name": "ncnr_DOI",
            "DOI": "10.18434/T4201B",
            "file_helper_url": "https://ncnr.nist.gov/ncnrdata/listncnrdatafiles_json.php"
        },
    ],
    
    # if not set, will instantiate all instruments.
    "instruments": ["refl", "ospec", "sans", "vsans"]
}



# This program is public domain
from copy import copy

import numpy as np

from reductus.dataflow.automod import cache, nocache, module, copy_module
# Note: do not load symbols from .steps directly into the file scope
# or they will be defined twice as reduction modules.
from . import steps

@cache
@module("candor")
def candor(
        filelist=None,
        bank_select=None,
        channel_select=None,
        dc_rate=0.,
        dc_slope=0.,
        detector_correction=False,
        monitor_correction=False,
        spectral_correction=False,
        attenuator_correction=False,
        Qz_basis='target',
        intent='auto',
        sample_width=None,
        base='auto'):
    r"""
    Load a list of Candor files from the NCNR data server.

    **Inputs**

    filelist (fileinfo[]): List of files to open.

    bank_select {Select bank} (int) : Choose bank to output (default is all)

    channel_select {Select channel(s)} (int[]*) : Choose channels to output from bank (default is all)

    dc_rate {Dark counts per minute} (float)
    : Number of dark counts to subtract from each detector channel per
    minute of counting time (see Dark Current).

    dc_slope {DC vs. slit 1} (float)
    : Dark counts per mm of slit 1 opening per minute.

    detector_correction {Apply detector deadtime correction} (bool)
    : If True, use deadtime constants in file to correct detector counts
    (see Detector Dead Time).

    monitor_correction {Apply monitor deadtime correction} (bool)
    : If True, use deadtime constants in file to correct monitor counts
    (see Monitor Dead Time).

    spectral_correction {Apply detector efficiency correction} (bool)
    : If True, scale counts by the detector efficiency calibration given
    in the file (see Spectral Efficiency).

    attenuator_correction {Apply attenuator corrections} (bool)
    : if True, scale the counts by the attenuator transmission coefficients given
    in the file, for the attenuator that was active for the data point.

    Qz_basis (opt:target|sample|detector|actual)
    : How to calculate Qz from instrument angles.

    intent (opt:auto|specular|background+\|background-\|intensity|rock sample|rock detector|rock qx|scan)
    : Measurement intent (specular, background+, background-, slit, rock),
    auto or infer.  If intent is 'scan', then use the first scanned variable.

    sample_width {Sample width (mm)} (float?)
    : Width of the sample along the beam direction in mm, used for
    calculating the effective resolution when the sample is smaller
    than the beam.  Leave blank to use value from data file.

    base {Normalize by} (opt:auto|monitor|time|roi|power|none)
    : How to convert from counts to count rates. Leave this as none if your
    template does normalization after integration (see Normalize).

    **Returns**

    output (candordata[]): All entries of all files in the list.

    | 2020-02-05 Paul Kienzle
    | 2020-03-12 Paul Kienzle Add slit 1 dependence for DC rate
    | 2020-08-27 Brian Maranville loader gets attenuator information
    | 2020-09-11 Brian Maranville loader updated with new motor names
    | 2020-11-18 Brian Maranville adding attenuator correction
    | 2020-11-20 Paul Kienzle support files with old detector table layout
    | 2021-06-11 David Hoogerheide compatibility with new DC rate module
    """
    from .load import url_load_list
    from .candor import load_entries
    auto_divergence = True

    # Note: candor automatically computes divergence.
    datasets = []
    for data in url_load_list(filelist, loader=load_entries):
        # TODO: drop data rows where fastShutter.openState is 0
        data.Qz_basis = Qz_basis
        if intent not in (None, 'none', 'auto'):
            data.intent = intent
        if bank_select is not None:
            data = select_bank(data, bank_select)
        if channel_select is not None:
            data = select_channel(data, channel_select)
        if auto_divergence:
            data = steps.divergence_fb(data, sample_width)
        if dc_rate != 0. or dc_slope != 0.:
            data = steps.dark_current([data], poly_coeff=[dc_slope, dc_rate])[0][0]       # now requires and returns a list of datasets; could move it out of the for loop
        if detector_correction:
            data = steps.detector_dead_time(data, None)
        if monitor_correction:
            data = steps.monitor_dead_time(data, None)
        if spectral_correction:
            data = spectral_efficiency(data)
        if attenuator_correction:
            data = attenuation(data)
        data = steps.normalize(data, base=base)
        datasets.append(data)

    return datasets

@module("candor")
def select_bank(data, bank_select=None):
    r"""
    Remove banks from Candor data.

    **Inputs**

    data (candordata): Input data with 1 or more banks

    bank_select {Select bank} (int) : Choose banks to output (default is all)

    **Returns**

    output (candordata): Data with indices not listed in bank_select removed.

    | 2020-07-07 Brian Maranville
    """

    if bank_select is not None:
        data = copy(data)
        data.detector.counts = data.detector.counts[:, :, [bank_select]]
        data.detector.counts_variance = data.detector.counts.copy()
        data.detector.dims = data.detector.counts.shape

        data.detector.efficiency = data.detector.efficiency[:, :, [bank_select]]
        data.detector.angle_x_offset = data.detector.angle_x_offset[[bank_select]]
        data.detector.wavelength = data.detector.wavelength[:, :, [bank_select]]
        data.detector.wavelength_resolution = data.detector.wavelength_resolution[:, :, [bank_select]]

    return data

@module("candor")
def select_channel(data, channel_select=None):
    r"""
    Select channels from bank of Candor data.

    **Inputs**

    data (candordata): Input data with 1 or more banks

    channel_select {Select channel(s)} (int[]) : Choose channels to output (default is all)

    **Returns**

    output (candordata): Data with indices not listed in channel_select removed.

    | 2020-07-12 Brian Maranville
    """

    if channel_select is not None:
        data = copy(data)
        data.detector = copy(data.detector)
        data.detector.counts = data.detector.counts[:, channel_select, :]
        data.detector.counts_variance = data.detector.counts_variance[:, channel_select, :]
        data.detector.dims = data.detector.counts.shape

        data.detector.efficiency = data.detector.efficiency[:, channel_select, :]
        data.detector.wavelength = data.detector.wavelength[:, channel_select, :]
        data.detector.wavelength_resolution = data.detector.wavelength_resolution[:, channel_select, :]
        if data._v is not None:
            data._v = data._v[:, channel_select, :]
        if data._dv is not None:
            data._dv = data._dv[:, channel_select, :]
    return data

@module("candor")
def select_points(data, point_select=None):
    r"""
    Select data points from scan of Candor.

    **Inputs**

    data (candordata): Input data with 1 or more banks

    point_select {Select points(s)} (int[]) : Choose points to output (default is all)

    **Returns**

    output (candordata): Data with point indices not listed in point_select removed.

    | 2020-07-03 Brian Maranville
    """

    if point_select is not None:
        data = copy(data)
        data.detector.counts = data.detector.counts[point_select]
        data.detector.counts_variance = data.detector.counts_variance[point_select]
        data.detector.dims = data.detector.counts.shape
        monitor = data.monitor
        monitor.counts = monitor.counts[point_select]
        monitor.counts_variance = monitor.counts_variance[point_select]
        monitor.count_time = monitor.count_time[point_select]
        monitor.roi_counts = monitor.roi_counts[point_select]
        monitor.roi_variance = monitor.roi_variance[point_select]
        monitor.source_power = monitor.source_power[point_select]
        monitor.source_power_variance = monitor.source_power_variance[point_select]

        if data._v is not None:
            data.v = data.v[point_select]
        if data._dv is not None:
            data.dv = data.dv[point_select]
        if data.angular_resolution is not None:
            data.angular_resolution = data.angular_resolution[point_select]

        for slit in (data.slit1, data.slit2, data.slit3, data.slit4):
            for attrname in ["x", "x_target", "y", "y_target"]:
                attr = getattr(slit, attrname, None)
                if attr is not None and isinstance(attr, np.ndarray):
                    setattr(slit, attrname, attr[point_select])

        data.sample.angle_x = data.sample.angle_x[point_select]
        data.sample.angle_x_target = data.sample.angle_x_target[point_select]
        data.detector.angle_x = data.detector.angle_x[point_select]
        data.detector.angle_x_target = data.detector.angle_x_target[point_select]

    return data


@module("candor")
def candor_select_xy(data, x_range=[None, None], y_range=[None, None]):
    r"""
    Select range of data from scan of Candor.

    **Inputs**

    data (candordata): Input data

    x_range (range?:x): x-region to keep from the data

    y_range (range?:y): y-region to keep from the data

    **Returns**

    output (candordata): Data with trimmed x and y values

    | 2020-10-14 Brian Maranville
    """

    if x_range is None and y_range is None:
        return data
    if x_range is None:
        x_range = [None, None]
    if y_range is None:
        y_range = [None, None]

    x_min, x_max = x_range
    y_min, y_max = y_range

    # y-axis is "points" axis, x-axis is "channels"
    (x, xlabel), (y, ylabel) = data.get_axes()
    x = x.flatten()
    y = y.flatten()

    x_select_min = x_min is None or x >= x_min
    x_select_max = x_max is None or x <= x_max
    x_select = np.logical_and(x_select_min, x_select_max)
    y_select_min = y_min is None or y >= y_min
    y_select_max = y_max is None or y <= y_max
    y_select = np.logical_and(y_select_min, y_select_max)
    #print('y', y, y_min, y_max, y_select, y.shape)
    #print('x', x, x_min, x_max, x_select, x.shape)

    if np.all(x_select):
        xdata = data # no limits on x
    else:
        channel_select = np.arange(x.shape[0], dtype="int")[x_select]
        #print('channel select:', channel_select)
        xdata = select_channel(data, channel_select)

    if np.all(y_select):
        xy_data = xdata
    else:
        point_select = np.arange(y.shape[0], dtype="int")[y_select]
        #print('point select:', point_select)
        xy_data = select_points(xdata, point_select)

    return xy_data


@module("candor")
def spectral_efficiency(data, spectrum=()):
    r"""
    Correct for the relative intensity in the different detector channels
    across the detector banks.  This correction depends on a number of
    factors including the distribution of wavelenths from the source,
    any wavelength selection filters in the path, the relative angles
    of the analyzer leaves, and the efficiency of the detector in each
    channel.

    **Inputs**

    data (candordata) : data to scale

    spectrum (float[]?) : override spectrum from data file

    **Returns**

    output (candordata) : scaled data

    | 2020-03-03 Paul Kienzle
    """
    #from .candor import NUM_CHANNELS
    # TODO: too many components operating directly on detector counts?
    # TODO: let the user paste their own spectral efficiency, overriding datafile
    # TODO: generalize to detector shapes beyond candor
    #print(data.v.shape, data.detector.efficiency.shape)
    #if len(spectrum)%NUM_CHANNELS != 0:
    #    raise ValueError("Vector length {s_len} must be a multiple of {NUM_CHANNELS}".format(s_len=len(spectrum), NUM_CHANNELS=NUM_CHANNELS))
    if spectrum:
        #spectrum = np.reshape(spectrum, (NUM_CHANNELS, -1)).T[None, :, :]
        spectrum = np.reshape(spectrum, data.detector.efficiency.shape)
    else:
        spectrum = data.detector.efficiency
    data = copy(data)
    data.detector = copy(data.detector)
    data.detector.counts = data.detector.counts / spectrum
    data.detector.counts_variance = data.detector.counts_variance / spectrum
    return data


@module("candor")
def recalibrate_wavelength(data, wavelengths=(), wavelength_resolutions=()):
    r"""
    Use different wavelength calibration from that already in the data file.

    **Inputs**

    data (candordata) : data to scale

    wavelengths (float[]?) : override wavelengths from data file

    wavelength_resolutions (float[]?) : override wavelength spreads (sigma) from data file

    **Returns**

    output (candordata) : scaled data

    | 2020-08-14 David Hoogerheide
    """
    # TODO: too many components operating directly on detector counts?
    # TODO: let the user paste their own spectral efficiency, overriding datafile
    # TODO: generalize to detector shapes beyond candor
    if wavelengths:
        wavelengths = np.reshape(wavelengths, data.detector.wavelength.shape)
    else:
        wavelengths = data.detector.wavelength
    if wavelength_resolutions:
        wavelength_resolutions = np.reshape(wavelength_resolutions, data.detector.wavelength_resolution.shape)
    else:
        wavelength_resolutions = data.detector.wavelength_resolution
    data = copy(data)
    data.detector = copy(data.detector)
    data.detector.wavelength = wavelengths
    data.detector.wavelength_resolution = wavelength_resolutions
    return data

@module("candor")
def attenuation(data, transmission=None, transmission_err=None, target_value=None):
    r"""
    Correct for wavelength-dependent attenuation from built-in attenuators

    Attenuation is specified per attenuator and detector,
    Attenuation_err should have the same shape, and is the width of the uncertainty for
    each transmission (1-sigma)

    If user-defined values are supplied, they will overwite the values in the data
    If no user-defined values are given, the values from the datafile will be used.

    **Inputs**

    data (candordata) : data to correct

    transmission (float[]?) : transmission vs. detector number (num_atten x num_detectors)

    transmission_err (float[]?) : 1-sigma width of uncertainty distribution of transmission

    target_value {Attenuator setting} (float[]?) : Attenuator setting, per point.  (npts)
        E.g. 1 if attenuator.key = 1 in NICE

    **Returns**

    corrected (candordata) : corrected data

    | 2020-08-24 Brian Maranville
    """

    from .candor import NUM_CHANNELS
    data = copy(data)
    data.attenuator = copy(data.attenuator)
    data.detector = copy(data.detector)

    if transmission is not None:
        data.attenuator.transmission = np.reshape(np.array(transmission), (-1, NUM_CHANNELS))
    if transmission_err is not None:
        data.attenuator.transmission_err = np.reshape(np.array(transmission_err), (-1, NUM_CHANNELS))
    if target_value is not None:
        data.attenuator.target_value = np.array(target_value)

    num_attenuators = data.attenuator.transmission.shape[0]

    for ai in range(1, num_attenuators+1):
        av = data.attenuator.target_value
        matching = (av == ai)
        if not np.any(matching):
            continue

        trans = data.attenuator.transmission[ai-1][None, :, None]
        trans_err = data.attenuator.transmission_err
        att = 1.0 / (trans)
        if trans_err is None or len(trans_err) < (ai-1):
            datt = np.zeros_like(att)
        else:
            datt = (trans_err[ai-1][None,:,None]) * att**2

        if data._v is not None:
            v = data._v[matching]
            if data._dv is not None:
                dv = data._dv[matching]
            else:
                dv = np.zeros_like(v)
            new_dv = np.sqrt(dv**2 * att**2 + datt**2 * v**2)
            data._dv[matching] = new_dv
            data._v[matching] *= att

        else:
            v = data.detector.counts[matching]
            var_v = data.detector.counts_variance[matching]
            new_var = var_v * att**2 + datt**2 * v**2
            data.detector.counts_variance[matching] = new_var
            data.detector.counts[matching] *= att

    return data

@module("candor")
def stitch_intensity(data, tol=0.001):
    r"""
    Join the intensity measurements into a single entry.

    **Inputs**

    data (candordata[]) : data to join

    tol {Tolerance (mm)} (float)
    : Tolerance for "overlapping" slit opening on slits 1, 2, and 3

    **Returns**

    output (candordata[]) : joined data

    | 2020-03-04 Paul Kienzle
    """
    from math import isclose
    from .refldata import Intent
    from .steps import join

    # sort the segments and make sure there is one and only one overlap
    data = [copy(v) for v in data]
    data.sort(key=lambda d: (d.slit1.x[0], d.slit2.x[0]))
    #print([(d.slit1.x[0],d.slit1.x[-1]) for d in data])
    for a, b in zip(data[:-1], data[1:]):
        # Verify overlap
        if (not isclose(a.slit1.x[-1], b.slit1.x[0], abs_tol=tol)
            or not isclose(a.slit2.x[-1], b.slit2.x[0], abs_tol=tol)
            or not isclose(a.slit3.x[-1], b.slit3.x[0], abs_tol=tol)):
            raise ValueError("need one point of overlap between segments")
        # Scale the *next* segment to the current segment rather than scaling
        # the current segment.  This does two things: (1) the first segment
        # doesn't need to be scaled since it has the narrowest slits, hence
        # the smallest attenuators are needed, and (2) the attenuation
        # computed at the next cycle automatically includes the cumulative
        # attenuation up to the current cycle.
        # Note: attenuators are wavelength-dependent, so keep separate
        # attenuation for each detector in each the bank.
        # TODO: propagate uncertainties
        # TODO: maybe update counts (unnormalized) rather than v (normalized)
        atten = a.v[-1] / b.v[0]
        b.v *= atten[None, ...]
        b.dv *= atten[None, ...]

        # Better scale estimation:
        # * user provides an overlap fitting radius
        # * order segments
        # * for each pair of segments
        # ** find midpoint between end of first segment and beginning of next segment
        # ** this midpoint may lie between the two segments (no overlap),
        #    exactly at the endpoints (single overlap) or somewhere
        #    before the end of the first and the beginning of the next
        # ** find all points within that overlap region
        #       low=[(s1_low, rate_low), ...], high=[(s1_high, rate_high), ...]
        #    where s1 is the slit 1 opening and rate is the normalized count rate.
        # ** define f(c) as chisq from fit a quadratic to the points
        #       [(s1_low, rate_low), ... , (s1_high, c*rate_high), ...]
        # ** minimize f(a, c, k) = a s1^2 + c - k y
        # This can be solved using a linear system:
        #      [a c k] = I
        # where
        #      a = [s1_low^2 s1_high^2]/sigma
        #      c = [1 ... 1  1 ... 1]/sigma
        #      k = [0 ... 0  y_high]/sigma
        #      I  = [y_low    0 ... 0]/sigma
        # With 1 point overlap and identical s1 can estimate k
        # With 2 point overlap can estimate a line bx + c
        # With 3 point overlap can estimate the quadratic as above
        # A quick simulation with 4 points overlap, a=5000, k=10
        # s1_low=[1,2,3,4], s1_high=[3,4,5,6] gave k to 0.5%
        # Note that the quadratic is centered at zero.  Without this condition
        # the quadratic can have a minimum over the range which is non-physical.
        # In practice the data is probably linear-quadratic-linear over the
        # range of the entire slit scan, but each overlap should be small
        # enough that it is either linear or quadratic.  With the right value
        # for "c" the curve can be made as flat as necessary over the range.
        # The attenuation "k" is not same across the detector bank since the
        # absorption coefficient is wavelength dependent and attenuation
        # is exponential with thickness of the attenuator.
        # However... we are going to have fixed attenuators on motor controls,
        # so its better to calibrate them once with high statistics and store
        # the attenuation values in the NeXus file.

    # Force intent to treat this as a slit scan.
    data[0].intent = Intent.slit

    # Use existing join algorithm.
    # TODO: expose join parameters to the user?
    data = join(data)
    return data


@module("candor")
def candor_rebin(data, qmin=None, qmax=None, qstep=0.003, qstep_max=None, average='gauss'):
    r"""
    Join the intensity measurements into a single entry.

    Early in the processing, when only normalized by monitor and time, points
    may be combined using Poisson statistics, where the uncertainty of the
    averaged rates will match the uncertainty in the rate if the measurement
    were performed in one step. As additional corrections are applied, such
    as deadtime corrections or other scaling this becomes less correct. After
    normalization by incident beam and more particularly, background
    subtraction, Poisson statistics are no longer reasonable and Gaussian
    statistics should be used to average the combined rates rather than
    recomputing the rates from the inferred counts.

    **Inputs**

    data (candordata) : data to join

    qmin (float) : Start of q range, or empty to infer from data

    qmax (float) : End of q range, or empty to infer from data

    qstep (float) : q step size at qmin, or 0 for no rebinning

    qstep_max (float) : maximum q step size at qmax, or empty to use qstep over entire q range

    average (opt:poisson|gauss) : combine bins using poisson or gaussian distribution

    **Returns**

    output (refldata) : joined data

    | 2020-03-04 Paul Kienzle
    | 2020-08-03 David Hoogerheide adding progressive q step coarsening
    | 2020-09-24 Brian Maranville changed default averaging
    | 2020-10-14 Paul Kienzle fixed uncertainty for time normalized data
    """
    from .candor import rebin, nobin

    if qstep == 0.0:
        data = nobin(data)
    else:
        if qmin is None:
            qmin = data.Qz.min()
        if qmax is None:
            qmax = data.Qz.max()
        if qstep_max is None:
            q = np.arange(qmin, qmax, qstep)
        else:
            dq = np.linspace(qstep,qstep_max, int(np.ceil(2*(qmax-qmin)/(qstep_max+qstep))))
            q = (qmin-qstep) + np.cumsum(dq)
        data = rebin(data, q, average)

    return data

candor_join = copy_module(
    steps.join, "candor_join",
    "refldata", "candordata", tag="candor")

candor_rescale = copy_module(
    steps.rescale, "candor_rescale",
    "refldata", "candordata", tag="candor")

candor_normalize = copy_module(
    steps.normalize, "candor_normalize",
    "refldata", "candordata", tag="candor")

candor_background = copy_module(
    steps.subtract_background, "candor_background",
    "refldata", "candordata", tag="candor")

candor_align = copy_module(
    steps.interpolate_background, "candor_align",
    "refldata", "candordata", tag="candor")

candor_divide = copy_module(
    steps.divide_intensity, "candor_divide",
    "refldata", "candordata", tag="candor")

candor_divergence_fb = copy_module(
    steps.divergence_fb, "candor_divergence",
    "refldata", "candordata", tag="candor")

candor_sample_broadening = copy_module(
    steps.sample_broadening, "candor_sample_broadening",
    "refldata", "candordata", tag="candor")



"""
Data smoothing

Uses moving window 1-D polynomial least squares smoothing filter.

"""
from __future__ import division

import numpy as np

from reductus.dataflow.lib.err1d import interp
from reductus.dataflow.lib.wsolve import smooth

def apply_smoothing(slits, dx, degree, span):
    assert span > degree, "Span must be greater than degree"
    s = np.hstack([[d.angular_resolution, d.slit1.x, d.slit2.x] for d in slits])
    s = find_common(s.T, dx=dx).T
    #import pylab;pylab.plot(np.arange(len(s[0])),s.T); pylab.show(); import sys; sys.exit()

    for d in slits:
        if span > 2:
            v, dv = smooth(s[0], d.angular_resolution, d.v, d.dv,
                           degree=degree, span=span)
        else:
            v, var = interp(s[0], d.angular_resolution, d.v, d.dv**2)
            dv = np.sqrt(var)
        d.slit1.x = s[1]
        d.slit2.x = s[2]
        d.v, d.dv = v, dv

def find_common(x, dx):
    x = x[np.argsort(x[:, 0])]
    xo, total, n = x[0]+0, x[0]+0, 1
    out = []
    for xk in x[1:]:
        if (xk[1:] - xo[1:] <= dx).all():
            total += xk
            n += 1
            continue
        out.append(total/n)
        xo, total, n = xk+0, xk+0, 1
    out.append(total/n)
    return np.vstack(out)



from reductus.dataflow import core as df
from reductus.dataflow.automod import make_modules

from . import steps
from . import templates
from .usansdata import RawData, USansData, USansCorData
from reductus.sansred.sansdata import Parameters

INSTRUMENT = "ncnr.usans"

def define_instrument():
    # Define modules
    modules = make_modules(steps.ALL_ACTIONS, prefix=INSTRUMENT+'.')

    # Define data types
    usansraw = df.DataType(INSTRUMENT+".raw", RawData)
    usansdata = df.DataType(INSTRUMENT+".data", USansData)
    usanscor = df.DataType(INSTRUMENT+".cor", USansCorData)
    params = df.DataType(INSTRUMENT+".params", Parameters)

    # Define instrument
    usans = df.Instrument(
        id=INSTRUMENT,
        name='NCNR Ultra-Small Angle Neutron Scattering Instrument (USANS)',
        menu=[('steps', modules)],
        datatypes=[usansdata, params, usanscor], # usansraw],
        template_defs=df.load_templates(templates),
        )

    # Register instrument
    df.register_instrument(usans)
    return usans



import os
import argparse

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--debug', action='store_true', help='autoload modules on change')
    parser.add_argument('-x', '--headless', action='store_true', help='do not automatically load client in browser')
    parser.add_argument('--external', action='store_true', help='listen on all interfaces, including external (local connections only if not set)')
    parser.add_argument('-p', '--port', default=8002, type=int, help='port on which to start the server')
    parser.add_argument('-c', '--config-file', type=str, help='path to JSON configuration to load')
    parser.add_argument('-i','--instruments', nargs='+', help='instruments to load (overrides config)')
    args = parser.parse_args()
    if args.config_file is not None:
        import json
        config = json.loads(open(args.config_file, 'rt').read())
    else:
        from reductus.dataflow.configure import load_config
        config = load_config(name="config", fallback=True)
    if args.instruments is not None:
        config["instruments"] = args.instruments
    # Strip "local" from data sources if running external
    if args.external:
        config["data_sources"] = [
            d for d in config["data_sources"]
            if d["name"] != "local"
        ]

    from reductus.web_gui.server_flask import create_app
    app = create_app(config)
    if not args.headless:
        import webbrowser
        webbrowser.open("http://localhost:%d" % (args.port))
    host = '0.0.0.0' if args.external else None
    app.run(port=args.port, host=host, debug=args.debug)

if __name__ == '__main__':
    main()



import sys

IS_PY3 = sys.version_info[0] >= 3

def _b(s):
    """string to byte conversion (utf-8)"""
    if IS_PY3:
        return s.encode('utf-8')
    else:
        return s

def _s(b):
    """byte to string conversion (utf-8)"""
    if IS_PY3:
        return b.decode('utf-8') if hasattr(b, 'decode') else b
    else:
        return b


from reductus.dataflow import core as df
from reductus.dataflow.automod import get_modules, make_modules, make_template

from . import magik_filters_func as steps
from . import templates
from .FilterableMetaArray import FilterableMetaArray

INSTRUMENT = "ncnr.ospec"

class Parameters(dict):
    def get_metadata(self):
        return self
    def get_plottable(self):
        return {"params": self.get_metadata(), "type": "params"}

def define_instrument():
    # Define modules
    actions = get_modules(steps)
    modules = make_modules(actions, prefix=INSTRUMENT+'.')

    # Define data types
    ospec2d = df.DataType(INSTRUMENT+".ospec2d", FilterableMetaArray)
    #ospec1d = df.DataType(INSTRUMENT+".ospec1d", FilterableMetaArray)
    ospecnd = df.DataType(INSTRUMENT+".ospecnd", FilterableMetaArray)
    params = df.DataType(INSTRUMENT+".params", Parameters)
    #offset_data = df.DataType(INSTRUMENT+".offset_data", dict)

    # Define instrument
    ospec = df.Instrument(
        id=INSTRUMENT,
        name='NCNR offspecular',
        menu=[('steps', modules)],
        datatypes=[ospec2d, ospecnd, params],
        template_defs=df.load_templates(templates),
        )

    # Register instrument
    df.register_instrument(ospec)
    return ospec


def loader_template():
    refl1d = df.lookup_instrument(INSTRUMENT)
    diagram = [
            ["ncnr_load", {}],
            ["divergence", {"data": "-.output"}],
    ]
    template = make_template(
        name="loader",
        description="loader only",
        diagram=diagram,
        instrument=refl1d,
        version=1.0,
        )
    return template


def unpolarized_template():
    refl1d = df.lookup_instrument(INSTRUMENT)
    diagram = [
        # Load the data
        ["ncnr_load", {}],

        # Preprocessing common to all data sets
        #["monitor_saturation", {"data": "-.output"}],
        #["detector_saturation", {"data": "-.output"}],
        ["divergence", {"data": "-.output"}],
        ["normalize", {"data": "-.output", "base": "auto"}],
        ["mark_intent", {"data": "-.output", "intent": ["auto"]}],
        ["group_by_intent => split", {"data": "-.output"}],

        # Preprocessing particular to each data type
        # Use nop to restart pipeline after split so that it is easier to
        # add/remove/rearrange steps for the different data types.
        ["nop", {"data": "split.specular"}],
        ["join => spec", {"data": "-.output"}],

        ["nop", {"data": "split.backp"}],
        ["mask_specular",  {"data": "-.output"}],
        ["align_background",  {"data": "-.output", "offset": "auto"}],
        ["join => backp",  {"data": "-.output"}],

        ["nop", {"data": "split.backm"}],
        ["mask_specular",  {"data": "-.output"}],
        ["align_background",  {"data": "-.output", "offset": "auto"}],
        ["join => backm",  {"data": "-.output"}],

        ["nop", {"data": "split.intensity"}],
        ["rescale",  {"data": "-.output", "scale": [1.0], "dscale": [0.0]}],
        ["join => intensity",  {"data": "-.output", "tolerance": 0.0001}],

        # Operate on the combined data for the final reduction
        ["subtract_background", {
            "data": "spec.output",
            "backp": "backp.output",
            "backm": "backm.output"}],
        ["divide_intensity",  {"data": "-.output", "base": "intensity.output"}],
        #["footprint",  {"data": "-.output")}],
    ]
    #for m in refl1d.modules: print m.__dict__
    #for m in refl1d.modules: print m.terminals[0]
    template = make_template(
        name="unpolarized",
        description="standard unpolarized reduction",
        diagram=diagram,
        instrument=refl1d,
        version=1.0,
        )
    return template

def demo():
    from reductus.dataflow.calc import process_template

    template = unpolarized_template()
    #print "========== Template ========"
    #template.show()
    #print "="*24
    test_dataset = [{'path': "ncnrdata/cgd/201511/21066/data/HMDSO_17nm_dry14.nxz.cgd",
                      'mtime': 1447353278}]
    refl = process_template(template=template,
                            config={"0": {"filelist": test_dataset}},
                            target=(len(template.modules)-1, "output"),
                            #target=(2, "data"),  # return an input
                            )
    #print "refl", refl.values
    return refl.values


if __name__ == "__main__":
    from reductus.dataflow.cache import use_redis
    use_redis()
    define_instrument()
    demo()




"""
Convert reduction flows to and from the stored representation.

The stored representation is given in terms of python primitives
which are suitable for calls to json loads/dumps.  The reduction
flows are represented using our internal instrument classes.  Only
the templates and the reduction flows need to be stored.


The following functions are available:

    dumps

        Convert reduction flow object to storable string.

    loads

        Convert storable string to reduction flow object.

"""
__all__ = ['dumps', 'loads']

import json

from . import core

def dumps(obj):
    if isinstance(obj, core.Template):
        classname = 'Template'
        version, state = obj.__getstate__()
    elif isinstance(obj, core.DataType):
        classname = 'DataType'
        version, state = obj.__getstate__()
    else:
        raise TypeError('unknown object "%s"' % str(type(obj)))

    return json.dumps([classname, version, state])

def loads(str):
    classname, version, state = json.loads(str)
    if classname == 'Template':
        obj = core.Template()
    elif classname == 'DataType':
        obj = core.DataType()
    else:
        raise TypeError('unknown object "%s"' % classname)
    obj.__setstate__((version, state))
    return obj

